from typing import Iterable, Mapping, Optional

import apteco_api as aa

from apteco.common import VariableType
from apteco.exceptions import get_deprecated_attr
from apteco.query import (
    ArrayClause,
    DateListClause,
    DateRangeClause,
    DateTimeRangeClause,
    FlagArrayClause,
    NPerVariableClause,
    NumericClause,
    SelectorClause,
    TextClause,
    general_error_msg_array,
    general_error_msg_date,
    general_error_msg_flag_array,
    general_error_msg_numeric,
    general_error_msg_selector,
    general_error_msg_text,
    normalize_date_input,
    normalize_date_value,
    normalize_datetime_value,
    normalize_number_input,
    normalize_number_value,
    normalize_string_input,
    normalize_string_value,
    single_value_error_msg_date,
    single_value_error_msg_datetime,
    single_value_error_msg_numeric,
    single_value_error_msg_text,
)


class Variable:
    """Class representing a FastStats system variable."""

    def __init__(
        self,
        name: str,
        description: str,
        type: str,
        folder_name: str,
        table: "Table",
        is_selectable: bool,
        is_browsable: bool,
        is_exportable: bool,
        is_virtual: bool,
        *,
        session: Optional["Session"] = None,
        **kwargs,
    ):
        self.name = name
        self.description = description
        self._model_type = type
        self.folder_name = folder_name
        self.table = table
        self.is_selectable = is_selectable
        self.is_browsable = is_browsable
        self.is_exportable = is_exportable
        self.is_virtual = is_virtual
        self.session = session

    @property
    def table_name(self):
        return self.table.name

    def _as_nper_clause(self, clause, n, by, ascending, label):
        return NPerVariableClause(
            clause=clause,
            n=n,
            per=self,
            by=by,
            ascending=ascending,
            label=label,
            session=self.session,
        )


class BaseSelectorVariable(Variable):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        selector_info = kwargs["selector_info"]  # type: aa.SelectorVariableInfo
        self.code_length = selector_info.code_length
        self.num_codes = selector_info.number_of_codes
        self.var_code_min_count = selector_info.minimum_var_code_count
        self.var_code_max_count = selector_info.maximum_var_code_count
        self.var_code_order = selector_info.var_code_order


class SelectorVariable(BaseSelectorVariable):
    """Class representing a FastStats Selector variable."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = VariableType.SELECTOR

    def __eq__(self, other):
        return SelectorClause(
            self,
            normalize_string_input(other, general_error_msg_selector),
            session=self.session,
        )

    def __ne__(self, other):
        return SelectorClause(
            self,
            normalize_string_input(other, general_error_msg_selector),
            include=False,
            session=self.session,
        )

    def _to_model_dimension(self):
        return aa.Dimension(id=self.name, type="Selector", variable_name=self.name)


class CombinedCategoriesVariable(BaseSelectorVariable):
    """Class representing a FastStats Combined Categories variable."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = "CombinedCategories"
        selector_info = kwargs["selector_info"]  # type: aa.SelectorVariableInfo
        self.combined_from = selector_info.combined_from_variable_name

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        raise NotImplementedError


class NumericVariable(Variable):
    """Class representing a FastStats Numeric variable."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = VariableType.NUMERIC
        numeric_info = kwargs["numeric_info"]  # type: aa.NumericVariableInfo
        self.min_value = numeric_info.minimum
        self.max_value = numeric_info.maximum
        self.is_currency = numeric_info.is_currency
        self.currency_locale = numeric_info.currency_locale
        self.currency_symbol = numeric_info.currency_symbol

    def __eq__(self, other):
        return NumericClause(
            self,
            normalize_number_input(other, general_error_msg_numeric),
            session=self.session,
        )

    def __ne__(self, other):
        return NumericClause(
            self,
            normalize_number_input(other, general_error_msg_numeric),
            include=False,
            session=self.session,
        )

    def __lt__(self, other):
        return NumericClause(
            self,
            [f"<{normalize_number_value(other, single_value_error_msg_numeric)}"],
            session=self.session,
        )

    def __le__(self, other):
        return NumericClause(
            self,
            [f"<={normalize_number_value(other, single_value_error_msg_numeric)}"],
            session=self.session,
        )

    def __gt__(self, other):
        return NumericClause(
            self,
            [f">{normalize_number_value(other, single_value_error_msg_numeric)}"],
            session=self.session,
        )

    def __ge__(self, other):
        return NumericClause(
            self,
            [f">={normalize_number_value(other, single_value_error_msg_numeric)}"],
            session=self.session,
        )

    def __getattr__(self, item):
        DEPRECATED_ATTRS = {
            "max": ("max_value", "0.7.0"),
            "min": ("min_value", "0.7.0"),
        }

        if item in DEPRECATED_ATTRS:
            return get_deprecated_attr(self, item, *DEPRECATED_ATTRS[item])
        else:
            raise AttributeError(
                f"'{self.__class__.__name__}' has no attribute '{item}'"
            )

    def missing(self, *, include=True, label=None):
        return NumericClause(
            self, ["><"], include=include, label=label, session=self.session
        )


class TextVariable(Variable):
    """Class representing a FastStats Text variable."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = VariableType.TEXT
        text_info = kwargs["text_info"]  # type: aa.TextVariableInfo
        self.max_length = text_info.maximum_text_length

    def equals(self, value, match_case=True, *, include=True, label=None):
        return TextClause(
            self,
            normalize_string_input(value, general_error_msg_text),
            match_type="Is",
            match_case=match_case,
            include=include,
            label=label,
            session=self.session,
        )

    def contains(self, value, match_case=True, *, include=True, label=None):
        return TextClause(
            self,
            normalize_string_input(value, general_error_msg_text),
            match_type="Contains",
            match_case=match_case,
            include=include,
            label=label,
            session=self.session,
        )

    def startswith(self, value, match_case=True, *, include=True, label=None):
        return TextClause(
            self,
            normalize_string_input(value, general_error_msg_text),
            match_type="Begins",
            match_case=match_case,
            include=include,
            label=label,
            session=self.session,
        )

    def endswith(self, value, match_case=True, *, include=True, label=None):
        return TextClause(
            self,
            normalize_string_input(value, general_error_msg_text),
            match_type="Ends",
            match_case=match_case,
            include=include,
            label=label,
            session=self.session,
        )

    def before(self, value, allow_equal=False, *, include=True, label=None):
        operator = "<=" if allow_equal else "<"
        normalized_value = normalize_string_value(value, single_value_error_msg_text)
        return TextClause(
            self,
            [f'{operator}"{normalized_value}"'],
            match_type="Ranges",
            match_case=False,
            include=include,
            label=label,
            session=self.session,
        )

    def after(self, value, allow_equal=False, *, include=True, label=None):
        operator = ">=" if allow_equal else ">"
        normalized_value = normalize_string_value(value, single_value_error_msg_text)
        return TextClause(
            self,
            [f'{operator}"{normalized_value}"'],
            match_type="Ranges",
            match_case=False,
            include=include,
            label=label,
            session=self.session,
        )

    def between(self, start, end, *, include=True, label=None):
        normalized_start = normalize_string_value(start, single_value_error_msg_text)
        normalized_end = normalize_string_value(end, single_value_error_msg_text)

        if normalized_start.lower() > normalized_end.lower():
            if normalized_start <= normalized_end:
                # inputted values were in correct order,
                # but the order is switched when lower-cased
                raise ValueError(
                    f"`start` must sort before `end`,"
                    f" but '{normalized_start}' sorts after '{normalized_end}'"
                    f" when compared case-insensitively."
                )
            raise ValueError("`start` must sort before `end`")

        return TextClause(
            self,
            [f'>="{normalized_start}" - <="{normalized_end}"'],
            match_type="Ranges",
            match_case=False,
            include=include,
            label=label,
            session=self.session,
        )

    def matches(self, value, match_case=True, *, include=True, label=None):
        normalized_input = normalize_string_input(value, general_error_msg_text)
        return TextClause(
            self,
            [f'="{v}"' for v in normalized_input],
            match_type="Ranges",
            match_case=match_case,
            include=include,
            label=label,
            session=self.session,
        )

    def __eq__(self, other):
        return self.equals(other, include=True)

    def __ne__(self, other):
        return self.equals(other, include=False)

    def __lt__(self, other):
        return self.before(other)

    def __le__(self, other):
        return self.before(other, allow_equal=True)

    def __gt__(self, other):
        return self.after(other)

    def __ge__(self, other):
        return self.after(other, allow_equal=True)


class ArrayVariable(BaseSelectorVariable):
    """Class representing a FastStats Array variable."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = VariableType.ARRAY

    def __eq__(self, other):
        return ArrayClause(
            self,
            normalize_string_input(other, general_error_msg_array),
            session=self.session,
        )

    def __ne__(self, other):
        return ArrayClause(
            self,
            normalize_string_input(other, general_error_msg_array),
            include=False,
            session=self.session,
        )


class FlagArrayVariable(BaseSelectorVariable):
    """Class representing a FastStats Flag Array variable."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = VariableType.FLAG_ARRAY

    def __eq__(self, other):
        return FlagArrayClause(
            self,
            normalize_string_input(other, general_error_msg_flag_array),
            session=self.session,
        )

    def __ne__(self, other):
        return FlagArrayClause(
            self,
            normalize_string_input(other, general_error_msg_flag_array),
            include=False,
            session=self.session,
        )


class BaseDateVariable(BaseSelectorVariable):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        selector_info = kwargs["selector_info"]  # type: aa.SelectorVariableInfo
        self.min_date = selector_info.minimum_date
        self.max_date = selector_info.maximum_date

        self.year = DateAccessor(self, "Years")
        self.quarter = DateAccessor(self, "Quarters")
        self.month = DateAccessor(self, "Months")
        self.day = DateAccessor(self, "Day")


class DateVariable(BaseDateVariable):
    """Class representing a FastStats Date variable."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = VariableType.DATE

    def __eq__(self, other):
        return DateListClause(
            self,
            normalize_date_input(other, general_error_msg_date, basic=True),
            session=self.session,
        )

    def __ne__(self, other):
        return DateListClause(
            self,
            normalize_date_input(other, general_error_msg_date, basic=True),
            include=False,
            session=self.session,
        )

    def __le__(self, other):
        return DateRangeClause(
            self,
            "Earliest",
            normalize_date_value(other, single_value_error_msg_date),
            session=self.session,
        )

    def __ge__(self, other):
        return DateRangeClause(
            self,
            normalize_date_value(other, single_value_error_msg_date),
            "Latest",
            session=self.session,
        )


class DateTimeVariable(BaseDateVariable):
    """Class representing a FastStats DateTime variable."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = VariableType.DATETIME

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        raise NotImplementedError

    def __le__(self, other):
        return DateTimeRangeClause(
            self,
            "Earliest",
            normalize_datetime_value(other, single_value_error_msg_datetime),
            session=self.session,
        )

    def __ge__(self, other):
        return DateTimeRangeClause(
            self,
            normalize_datetime_value(other, single_value_error_msg_datetime),
            "Latest",
            session=self.session,
        )


class ReferenceVariable(Variable):
    """Class representing a FastStats Reference variable."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = VariableType.REFERENCE

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        raise NotImplementedError


class VariablesAccessor:
    """List-like and dictionary-like access for variables."""

    def __init__(self, variables: Iterable[Variable]):
        self._variables = list(variables)
        self._variables_by_name = {var.name: var for var in self._variables}
        self._variables_by_desc = {var.description: var for var in self._variables}
        self.names = VariableNamesAccessor(self._variables_by_name)
        self.descs = VariableDescsAccessor(self._variables_by_desc)

    @property
    def descriptions(self):
        """Alias of ``descs``"""
        return self.descs

    def __len__(self):
        return len(self._variables)

    def __iter__(self):
        return iter(self._variables)

    def __getitem__(self, item):
        name_match = self._variables_by_name.get(item)
        desc_match = self._variables_by_desc.get(item)
        match_count = (name_match is not None) + (desc_match is not None)
        if match_count == 1:
            return name_match or desc_match
        elif match_count == 2:
            if name_match.name == desc_match.name:
                return name_match
            raise KeyError(f"Lookup key '{item}' was ambiguous.") from None
        else:
            raise KeyError(
                f"Lookup key '{item}' did not match a variable name or description."
            ) from None


class VariableNamesAccessor:
    """Dictionary-like access for variables by name."""

    def __init__(self, variables_by_name: Mapping[str, Variable]):
        self._variables_by_name = dict(variables_by_name)
        self._variable_names = list(self._variables_by_name.keys())

    def __iter__(self):
        return iter(self._variable_names)

    def __getitem__(self, item):
        try:
            return self._variables_by_name[item]
        except KeyError:
            raise KeyError(
                f"Lookup key '{item}' did not match a variable name."
            ) from None


class VariableDescsAccessor:
    """Dictionary-like access for variables by description."""

    def __init__(self, variables_by_desc: Mapping[str, Variable]):
        self._variables_by_desc = dict(variables_by_desc)
        self._variable_descs = list(self._variables_by_desc.keys())

    def __iter__(self):
        return iter(self._variable_descs)

    def __getitem__(self, item):
        try:
            return self._variables_by_desc[item]
        except KeyError as exc:
            raise KeyError(
                f"Lookup key '{item}' did not match a variable description."
            ) from None


class DateAccessor:
    """Banding for date variables."""

    def __init__(self, variable, banding):
        self.variable = variable
        self.table = variable.table
        self.banding = banding
        self.type = VariableType.BANDED_DATE
        self.name = f"{variable.name}_{banding.rstrip('s')}"
        self.description = f"{variable.description} ({banding.rstrip('s')})"

    def _to_model_dimension(self):
        return aa.Dimension(
            id=self.name,
            type="DateBand",
            variable_name=self.variable.name,
            banding=aa.DimensionBanding(self.banding),
        )



"""Exceptions raised and warnings issued by the py-apteco package."""

import sys
import warnings


class AptecoException(Exception):
    """Base exception class for all exceptions in the package."""


class ApiResultsError(AptecoException):
    """Raised when API returns inconsistent or unexpected results."""


class TablesError(AptecoException):
    """Raised when error occurs initialising tables data or tree."""


class VariablesError(AptecoException):
    """Raised when error occurs initialising variables."""


class DeserializeError(AptecoException):
    """Raised when error occurs when deserializing objects."""


class AptecoWarning(Warning):
    """Base class for all warnings in the package."""


# https://docs.python.org/3.7/whatsnew/3.7.html#whatsnew37-pep565
if sys.version_info[:2] < (3, 7):
    deprecation_warning_class = FutureWarning
else:
    deprecation_warning_class = DeprecationWarning


class AptecoDeprecationWarning(deprecation_warning_class, AptecoWarning):
    """Warning about deprecated features."""


def get_deprecated_attr(obj, old_name, new_name, version_deprecated, stacklevel=3):
    """Issue warning for accessing deprecated attribute."""
    warnings.warn(
        f"'{old_name}' is deprecated since py-apteco v{version_deprecated},"
        f" use '{new_name}' instead",
        AptecoDeprecationWarning,
        stacklevel=stacklevel,
    )
    return getattr(obj, new_name)



import itertools

import apteco_api as aa
import numpy as np
import pandas as pd

from apteco.common import VariableType


class Cube:
    def __init__(
        self, dimensions, measures=None, selection=None, table=None, *, session=None
    ):
        self.dimensions = dimensions
        self.measures = measures
        self.selection = selection
        self.table = table
        self.session = session
        self._check_inputs()
        self._data, self._sizes, self._headers, self._measure_names = self._get_data()

    def _check_inputs(self):
        if self.session is None:
            raise ValueError("You must provide a valid session (none was given).")
        if not self.dimensions:
            raise ValueError(
                "You must specify at least one variable"
                " to use as a dimension on the cube (none was given)."
            )
        self._check_table()
        self._check_dimensions()
        self._check_measures()
        self._check_relations()

    def _check_table(self):
        if self.table is None:
            if self.selection is None:
                raise ValueError(
                    "You must specify the resolve table of the cube"
                    " (no table or selection was given)."
                )
            self.table = self.selection.table

    def _check_dimensions(self):
        for dimension in self.dimensions:
            if dimension.type not in (VariableType.SELECTOR, VariableType.BANDED_DATE):
                raise ValueError(
                    f"The variable '{dimension.name}' has type '{dimension.type}'."
                    f"\nOnly Selector variables (excluding sub-types)"
                    f" and banded Date variables"
                    f" are currently supported as cube dimensions."
                )
            if not dimension.table.is_related(self.table, allow_same=True):
                raise ValueError(
                    f"The resolve table of the cube is '{self.table.name}',"
                    f" but the variable '{dimension.name}' belongs to the"
                    f" '{dimension.table.name}' table."
                    f"\nOnly variables from the same table as the cube"
                    f" or from related tables can be used as cube dimensions."
                )

    def _check_measures(self):
        if self.measures is None:
            self.measures = [self.table]
            return
        for measure in self.measures:
            if not hasattr(measure, "_to_model_measure"):
                raise ValueError(f"Invalid measure given: must be statistic")
            if not measure.table.is_related(self.table, allow_same=True):
                raise ValueError(
                    f"The resolve table of the cube is '{self.table.name}',"
                    f" but the measure '{measure._name}' belongs to the"
                    f" '{measure.table.name}' table."
                    f"\nOnly measures from the same table as the cube"
                    f" or from related tables can be used as cube dimensions."
                )

    def _check_relations(self):
        d_elements = [
            (d, f"dimension '{d.name}' (table: {d.table.name})")
            for d in self.dimensions
        ]
        m_elements = [(m, f"measure '{m._name}' (table: {m.table.name})") for m in self.measures]

        cross_cube = False
        non_related = []
        for x, y in itertools.combinations(d_elements, r=2):
            if x[0].table.is_related(y[0].table, allow_same=True):
                pass
            elif x[0].table.is_descendant(self.table) and y[0].table.is_descendant(self.table):
                cross_cube = True
            else:
                # should be unreachable
                non_related.append((x, y))

        if non_related:
            error_msg = (
                "Dimension tables must either be related,"
                " or be descendants of the resolve table."
                " The following pair(s) of dimensions did not meet those criteria:"
            )
            for x, y in non_related:
                error_msg += f"\n{x[1]} & {y[1]}"
            raise ValueError(error_msg)

        if cross_cube:
            cell_table = self.table
            bad_measures = []
            for m in m_elements:
                if not m[0].table.is_ancestor(cell_table, allow_same=True):
                    bad_measures.append(m)
            if bad_measures:
                error_msg = (
                    "For 'cross cubes' (where two dimensions are unrelated),"
                    " all measures must be from the resolve table"
                    " or one of its ancestors."
                    " The following measure(s) did not meet those criteria:"
                )
                for bm in bad_measures:
                    error_msg += f"\n{bm[1]}"
                raise ValueError(error_msg)

        else:
            # cell_table = max(dim.table for dim in self.dimensions)
            non_related = []
            for m, d in itertools.product(m_elements, d_elements):
                if not m[0].table.is_related(d[0].table, allow_same=True):
                    non_related.append((m, d))
            if non_related:
                error_msg = (
                    "Measure and dimension tables must be related."
                    " The following measure-dimension pair(s) "
                    " did not meet those criteria:"
                )
                for m, d in non_related:
                    error_msg += f"\n{m[1]} & {d[1]}"
                raise ValueError(error_msg)

    def _get_data(self):
        cube_result = self._get_cube()
        raw_data = [
            [x for row in mr.rows for x in row.split("\t")]
            for mr in cube_result.measure_results
        ]
        headers = [
            {
                "codes": [
                    "TOTAL" if c == "iTOTAL" else c
                    for c in dimension.header_codes.split("\t")
                ],
                "descs": [
                    "TOTAL" if d == "iTOTAL" else d
                    for d in dimension.header_descriptions.split("\t")
                ],
            }
            for dimension in reversed(cube_result.dimension_results)
        ]
        sizes = tuple(len(dh["codes"]) for dh in headers)
        data_as_arrays = [np.array(raw_measure_data) for raw_measure_data in raw_data]
        data = [
            measure_data_as_array.reshape(sizes)
            for measure_data_as_array in data_as_arrays
        ]
        measure_names = [mr.id for mr in cube_result.measure_results]
        dimension_names = [dim.id for dim in cube_result.dimension_results]
        return data, sizes, headers, measure_names

    def _get_cube(self):
        cube = aa.Cube(
            base_query=aa.Query(
                selection=self.selection._to_model_selection()
                if self.selection is not None
                else aa.Selection(table_name=self.table.name)
            ),
            resolve_table_name=self.table.name,
            storage="Full",
            dimensions=self._create_dimensions(),
            measures=self._create_measures(),
        )
        cubes_controller = aa.CubesApi(self.session.api_client)
        cube_result = cubes_controller.cubes_calculate_cube_synchronously(
            self.session.data_view, self.session.system, cube=cube
        )
        return cube_result

    def _create_dimensions(self):
        return [d._to_model_dimension() for d in reversed(self.dimensions)]

    def _create_measures(self):
        return [m._to_model_measure(self.table) for m in self.measures]

    def to_df(
        self, unclassified=False, totals=False, no_trans=False, convert_index=None
    ):
        # 0. validate inputs
        if no_trans is not False:  # no_trans not currently supported
            raise ValueError("no_trans must be False")

        if any([unclassified, totals, no_trans]):
            if convert_index is True:
                raise ValueError(
                    "Cannot convert index if any of unclassified, totals, no_trans"
                    " is included in dataframe"
                )
            elif convert_index is None:
                convert_index = False
        else:
            if convert_index is None:
                convert_index = True

        # 1. normalize headers
        normalized_headers = [
            self._normalize_headers(headers, dimension)
            for headers, dimension in zip(self._headers, self.dimensions)
        ]

        # 2. create slices for filtering
        start = 1 if not unclassified else 0
        end = -1 if not totals else None
        slices = [slice(start, end) for __ in self.dimensions]

        # 3. apply slices
        data = [measure_data[tuple(slices)] for measure_data in self._data]
        sliced_headers = [headers[s] for headers, s in zip(normalized_headers, slices)]

        # 4. convert headers
        if convert_index:
            converted_headers = [
                self._convert_headers(headers, dimension)
                for headers, dimension in zip(sliced_headers, self.dimensions)
            ]
        else:
            converted_headers = sliced_headers

        # 5. create index
        if len(self.dimensions) == 1:
            index = pd.Index(converted_headers[0], name=self.dimensions[0].description)
        else:
            index = pd.MultiIndex.from_product(
                converted_headers, names=[d.description for d in self.dimensions]
            )

        # 6. convert data & create DataFrame
        return pd.DataFrame(
            {
                measure_name: pd.to_numeric(measure_data.ravel(), errors="coerce")
                for measure_name, measure_data in zip(self._measure_names, data)
            },
            index=index,
        )

    @staticmethod
    def _normalize_headers(headers, dimension):
        variable_type = dimension.type
        if variable_type == VariableType.SELECTOR:
            return headers["descs"]
        elif variable_type == VariableType.BANDED_DATE:

            normalizer = {
                "Years": None,  # "%Y"
                "Quarters": None,  # "%YQ{q}"
                "Months": lambda x: f"{x[0:4]}-{x[4:6]}",  # "%Y%m"->"%Y-%m"
                "Day": lambda x: f"{x[0:4]}-{x[4:6]}-{x[6:8]}",  # "%Y%m%d"->"%Y-%m-%d"
            }[dimension.banding]

            if normalizer is None:
                normalized = headers["codes"][1:-1]
            else:
                normalized = [normalizer(h) for h in headers["codes"][1:-1]]

            unclassified = {
                "Years": "0000",
                "Quarters": "000000",
                "Months": "000000",
                "Day": "00000000",
            }[dimension.banding]

            assert headers["codes"][0] == unclassified
            assert headers["codes"][-1] == "TOTAL"

            return ["Unclassified"] + normalized + ["TOTAL"]

        else:
            raise ValueError(f"Unrecognised dimension type: {variable_type}")

    @staticmethod
    def _convert_headers(headers, dimension):
        variable_type = dimension.type
        if variable_type == VariableType.SELECTOR:
            return headers
        elif variable_type == VariableType.BANDED_DATE:
            period = {"Years": "Y", "Quarters": "Q", "Months": "M", "Day": "D"}[
                dimension.banding
            ]
            return pd.PeriodIndex(headers, freq=period)
        else:
            raise ValueError(f"Unrecognised dimension type: {dimension}")



from datetime import date, datetime
from decimal import Decimal
from fractions import Fraction
from numbers import Integral, Number, Rational, Real
from typing import Iterable, List, Optional

import apteco_api as aa

from apteco.common import VariableType
from apteco.cube import Cube
from apteco.datagrid import DataGrid
from apteco.exceptions import AptecoException

DECIMAL_PLACES = 4


class OperationError(AptecoException):
    """Raised when invalid operation attempted on clauses."""


class InvalidValuesError(AptecoException):
    """Raised when invalid values used in selection query."""


class SelectionCount(object):
    def __init__(self, count_response):
        self.table_name = count_response.table_name
        self.count = count_response.count_value


class Selection:
    def __init__(self, query: aa.Query, session: "Session"):

        self.queries_controller = aa.QueriesApi(session.api_client)
        self._response = self._run_query(session.data_view, session.system, query)

        self.counts = []
        self.count_names = []
        for count_dict in self._response.counts:
            new_count = SelectionCount(count_dict)
            self.counts.append(new_count)
            self.count_names.append(new_count.table_name)

        self.count = self.counts[0].count
        self.table_name = self.counts[0].table_name

    def _run_query(self, data_view_name: str, system: str, query: aa.Query):
        """Request a query be counted

        Args:
            data_view_name: dataView that the session is logged in to
            system: system with query
            query: query request

        Returns:
            QuerySimpleResult, unused kwargs

        """

        response = self.queries_controller.queries_perform_query_count_synchronously(
            data_view_name=data_view_name, system_name=system, query=query
        )

        return response  # type: aa.QueryResult

    def get_count(self, table: str = None) -> int:
        """Get the count for the given table.

        If you pass this with no argument,
        it will return the count of the current resolve table
        (so ``selection.get_count()`` is the same as
        ``selection.count``).

        Args:
            table: the name of the table

        Returns:
            int: the number of records in the selection
            for the given table

        """

        if table is None:
            return self.count
        elif table in self.count_names:
            return self.counts[self.count_names.index(table)].count
        else:
            raise ValueError("Table name not found")

    def set_table(self, table: str) -> None:
        """Set the resolve table level for the selection.

        This changes ``count`` and ``table_name``
        correspondingly.
        This is only available for any tables included
        in ancestor table counts.

        Args:
            table: the name of the table

        """

        if table in self.count_names:
            count_idx = self.count_names.index(table)
            self.count = self.counts[count_idx].count
            self.table_name = self.counts[count_idx].table_name
        else:
            raise ValueError("Table name not found")


# TODO: change name to Controller (confusing with APIClient otherwise)
class APIController:
    def __init__(self, controller, session: "Session"):
        self._controller = controller(session.api_client)

    def __call__(self, function_as_string, **kwargs):
        return getattr(self._controller, function_as_string)(**kwargs)


class TableMixin:
    def count(self):
        query_final = aa.Query(
            selection=aa.Selection(table_name=self.name, ancestor_counts=True)
        )
        session = self.session
        return Selection(query_final, session).count


general_error_msg_selector = (
    "Chosen value(s) for a selector variable"
    " must be given as a string or an iterable of strings."
)
general_error_msg_text = (
    "Chosen value(s) for a text variable"
    " must be given as a string or an iterable of strings."
)
single_value_error_msg_text = "Must specify a single string for this type of operation."
general_error_msg_array = (
    "Chosen value(s) for an array variable"
    " must be given as a string or an iterable of strings."
)
general_error_msg_flag_array = (
    "Chosen value(s) for a flag array variable"
    " must be given as a string or an iterable of strings."
)


def normalize_string_value(value, error_msg):
    if isinstance(value, str):
        return value
    else:
        raise ValueError(error_msg)


def normalize_string_input(input_value, error_msg):
    if isinstance(input_value, str):
        return [normalize_string_value(input_value, error_msg)]
    elif isinstance(input_value, Iterable):
        return [normalize_string_value(v, error_msg) for v in input_value]
    else:
        raise ValueError(error_msg)


general_error_msg_numeric = (
    "Chosen value(s) for a numeric variable"
    " must be given as a number or an iterable of numbers."
)
single_value_error_msg_numeric = (
    "Must specify a single number for this type of operation."
)


def normalize_number_value(value, error_msg):
    if isinstance(value, (Real, Decimal)) and not isinstance(value, bool):
        if isinstance(value, Integral):
            return str(int(value))
        if isinstance(value, Decimal):
            return str(value.quantize(Decimal(10) ** -DECIMAL_PLACES))
        if isinstance(value, (Real, Rational)):
            return str(round(float(value), DECIMAL_PLACES))
    else:
        raise ValueError(error_msg)


def normalize_number_input(input_value, error_msg):
    if isinstance(input_value, Number):
        return [normalize_number_value(input_value, error_msg)]
    elif isinstance(input_value, Iterable) and not isinstance(input_value, str):
        return [normalize_number_value(v, error_msg) for v in input_value]
    else:
        raise ValueError(error_msg)


general_error_msg_date = (
    "Chosen value for a date variable"
    " must be a date object or an iterable of date objects."
)
single_value_error_msg_date = "Must specify a single date for this type of operation."


def normalize_date_value(value, error_msg, *, basic=False):
    if isinstance(value, date):
        if basic:  # format for DateListClause
            return value.strftime("%Y%m%d")
        else:
            return value.strftime("%Y-%m-%d")
    else:
        raise ValueError(error_msg)


def normalize_date_input(input_value, error_msg, *, basic=False):
    if isinstance(input_value, date):
        return [normalize_date_value(input_value, error_msg, basic=basic)]
    elif isinstance(input_value, Iterable) and not isinstance(input_value, str):
        return [normalize_date_value(v, error_msg, basic=basic) for v in input_value]
    else:
        raise ValueError(error_msg)


single_value_error_msg_datetime = (
    "Must specify a single datetime for this type of operation."
)


def normalize_datetime_value(value, error_msg):
    if isinstance(value, datetime):
        return value.isoformat(timespec="seconds")  # "%Y-%m-%dT%H:%M:%S"
    else:
        raise ValueError(error_msg)


def normalize_datetime_input(input_value, error_msg):
    if isinstance(input_value, datetime):
        return [normalize_datetime_value(input_value, error_msg)]
    elif isinstance(input_value, Iterable) and not isinstance(input_value, str):
        return [normalize_datetime_value(v, error_msg) for v in input_value]
    else:
        raise ValueError(error_msg)


class Clause:
    @property
    def table_name(self):
        return self.table.name

    def count(self):
        query_final = aa.Query(selection=self._to_model_selection())
        session = self.session
        return Selection(query_final, session).count

    def _to_model_selection(self):
        return aa.Selection(
            table_name=self.table_name,
            ancestor_counts=True,
            rule=aa.Rule(clause=self._to_model_clause()),
        )

    def _change_table(self, new_table, simplify=False):

        if simplify:
            return self._strip_table_changes()._change_table_main(new_table)
        else:
            return self._change_table_main(new_table)

    def _change_table_main(self, new_table):
        if self.table.is_descendant(new_table):
            if self.table.parent == new_table:
                return logic_clause("ANY", [self], new_table)
            else:
                return logic_clause(
                    "ANY", [self], self.table.parent
                )._change_table_main(new_table)
        elif self.table.is_ancestor(new_table):
            if self.table == new_table.parent:
                return logic_clause("THE", [self], new_table)
            else:
                return logic_clause(
                    "THE", [self._change_table_main(new_table.parent)], new_table
                )
        elif self.table == new_table:
            return self
        else:
            try:
                common_ancestor_names = set(t.name for t in self.table.ancestors) & set(
                    t.name for t in new_table.ancestors
                )
                nearest_common_ancestor = max(
                    self.session.tables[name] for name in common_ancestor_names
                )
            except ValueError:  # empty list passed to max() -> no common ancestors
                raise OperationError(
                    f"Could not establish relationship between tables "
                    f"'{new_table.name}' and '{self.table_name}'."
                ) from None
            return self._change_table_main(nearest_common_ancestor)._change_table_main(
                new_table
            )

    def _strip_table_changes(self):
        if self.logic is None or self.logic.operation.upper() in ["AND", "OR", "NOT"]:
            return self
        elif self.logic.operation.upper() in ["ANY", "THE"]:
            if len(self.logic.operands) != 1:
                raise OperationError(
                    f"'{self.logic.operation.upper()}' has invalid "
                    f"number of operands ({len(self.logic.operands)}). "
                    f"Should be 1."
                )
            return self.logic.operands[0]._strip_table_changes()
        else:
            raise OperationError(
                f"Encountered unexpected logic clause "
                f"'{self.logic.operation.upper()}'."
            )

    def __and__(self, other):
        if self.table != other.table:
            return logic_clause("AND", [self, other._change_table(self.table)])
        return logic_clause("AND", [self, other])

    def __or__(self, other):
        if self.table != other.table:
            return logic_clause("OR", [self, other._change_table(self.table)])
        return logic_clause("OR", [self, other])

    def __invert__(self):
        return logic_clause("NOT", [self])

    def __mul__(self, other):
        return self._change_table(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def datagrid(self, columns, table=None, max_rows=1000):
        return DataGrid(
            columns,
            selection=self,
            table=table if table is not None else self.table,
            max_rows=max_rows,
            session=self.session,
        )

    def cube(self, dimensions, measures=None, table=None):
        return Cube(
            dimensions,
            measures=measures,
            selection=self,
            table=table if table is not None else self.table,
            session=self.session,
        )

    def sample(
        self, n=None, frac=None, sample_type="Random", skip_first=0, *, label=None
    ):
        if sum((i is not None) for i in (n, frac)) != 1:
            raise ValueError("Must specify either n or frac")
        total = None
        percent = None
        fraction = None
        if n is not None:
            if not isinstance(n, Integral) or n < 1:
                raise ValueError("n must be an integer greater than 0")
            total = int(n)
        else:
            if not isinstance(frac, Real):
                raise ValueError("frac must be either a float or a fraction")
            if not 0 < float(frac) < 1:
                raise ValueError("frac must be between 0 and 1")
            if isinstance(frac, Rational):
                fraction = Fraction(frac.numerator, frac.denominator)
            else:
                percent = float(frac) * 100

        if sample_type.title() not in ("First", "Stratified", "Random"):
            raise ValueError(f"'{sample_type}' is not a valid sample type")
        sample_type = sample_type.title()

        if not isinstance(skip_first, Integral) or int(skip_first) < 0:
            raise ValueError("`skip_first` must be a non-negative integer")
        skip_first = int(skip_first)

        return LimitClause(
            clause=self,
            total=total,
            percent=percent,
            fraction=fraction,
            sample_type=sample_type,
            skip_first=skip_first,
            label=label,
            session=self.session,
        )

    def limit(
        self, n=None, frac=None, by=None, ascending=None, per=None, *, label=None
    ):
        if ascending is not None:
            if not isinstance(ascending, bool):
                raise ValueError("ascending must be True or False")
            if by is None:
                raise ValueError("Must specify by with ascending")
        else:
            ascending = False
        # if by is not None and not isinstance(by, Variable):
        #     raise ValueError("by must be a variable")
        if per is not None:
            # nper
            try:
                return per._as_nper_clause(
                    clause=self, n=n, by=by, ascending=ascending, label=label
                )
            except AttributeError:
                raise ValueError("`per` must be a table or a variable") from None
        if by is None:
            # call.sample()
            return self.sample(n=n, frac=frac, label=label)
        # topn
        if n is not None:
            return TopNClause(
                clause=self,
                total=n,
                percent=None,
                by=by,
                ascending=ascending,
                label=label,
                session=self.session,
            )
        if frac is not None:
            if isinstance(frac, tuple):
                percent = tuple(i * 100 for i in frac)
            elif isinstance(frac, Real):
                percent = frac * 100
            else:
                raise ValueError()
            return TopNClause(
                clause=self,
                total=None,
                percent=percent,
                by=by,
                ascending=ascending,
                label=label,
                session=self.session,
            )
        raise ValueError()


class SelectorClauseMixin:
    @staticmethod
    def _verify_values(values: Iterable[str], variable_name: str, session: "Session"):
        systems_controller = APIController(aa.FastStatsSystemsApi, session)

        # TODO: separate this part into a (dynamic) getter for a .values property
        #  on the SelectorVariable class
        var_codes_raw = []
        offset = 0
        while True:
            var_code_results = systems_controller(
                "fast_stats_systems_get_fast_stats_variable_codes",
                data_view_name=session.data_view,
                system_name=session.system,
                variable_name=variable_name,
                offset=offset,
            )
            var_codes_raw.extend(var_code_results.list)
            if (
                var_code_results.offset + var_code_results.count
                >= var_code_results.total_count
            ):
                break
            offset = var_code_results.offset + var_code_results.count

        values_by_code = {c.code: c for c in var_codes_raw}
        values_by_desc = {c.description: c for c in var_codes_raw}

        val_set = set(values)
        code_set = set(values_by_code.keys())
        desc_set = set(values_by_desc.keys())
        if val_set <= code_set:
            return sorted(val_set)
        elif val_set <= desc_set:
            return [values_by_desc[desc].code for desc in sorted(val_set)]
        elif val_set <= (code_set | desc_set):
            raise InvalidValuesError(
                "Cannot mix codes and descriptions in selector values list."
            )
        else:
            valid_set = val_set & (code_set | desc_set)
            invalid_set = val_set - (code_set | desc_set)
            invalid_size = len(invalid_set)
            sample_size = min(invalid_size, 3)
            invalid_sample = [f"'{invalid_set.pop()}'" for _ in range(sample_size)]
            sample_msg = f" (sample of {sample_size})" if invalid_size > 3 else ""
            sample_str = ", ".join(invalid_sample) + sample_msg
            if valid_set <= code_set:
                raise InvalidValuesError(
                    f"{invalid_size} invalid code(s) "
                    f"detected in selector values list:\n{sample_str}"
                )
            elif valid_set <= desc_set:
                raise InvalidValuesError(
                    f"{invalid_size} invalid description(s) "
                    f"detected in selector values list:\n{sample_str}"
                )
            elif valid_set <= code_set | desc_set:
                raise InvalidValuesError(
                    f"{invalid_size} invalid value(s) "
                    f"detected in selector values list. "
                    f"Also, cannot mix codes and descriptions "
                    f"in selector values list:\n{sample_str}"
                )
            else:
                raise InvalidValuesError(
                    "Selector values list contains no valid values."
                )


def logic_clause(
    operation: str,
    operands: List["Clause"],
    new_table: Optional["Table"] = None,
    *,
    name: Optional[str] = None,
):
    if operation in ["AND", "OR", "NOT"]:
        return BooleanClause(
            operands[0].table,
            operation,
            operands,
            label=name,
            session=operands[0].session,
        )
    elif operation in ["ANY", "THE"]:
        (operand,) = operands
        return TableClause(
            new_table, operation, operand, label=name, session=operand.session
        )
    else:
        valid_operations = ["ANY", "AND", "OR", "NOT", "THE"]
        raise ValueError(
            f"'{operation}' is not a valid operation: "
            f"must be one of: {', '.join(valid_operations)}"
        )


class LogicClauseMixin:
    @staticmethod
    def _validate_operation(operation, operands, new_table):

        if len(operands) == 0:
            raise OperationError("The operands list must contain at least one clause.")
        if operation in ["ANY", "NOT", "THE"] and len(operands) != 1:
            raise OperationError(
                f"'{operation}' can only be performed on a single clause."
            )
        if operation in ["AND", "OR"]:
            if len(operands) == 1:
                raise OperationError(
                    f"'{operation}' requires the operands list to contain "
                    f"at least two clauses."
                )
            if new_table is not None:
                raise OperationError(
                    f"Cannot change tables while performing '{operation}'."
                )

        if operation == "NOT":
            return operands[0]._table

        if operation in ["ANY", "THE"]:
            if new_table is None:
                raise OperationError(
                    f"'{operation}' requires "
                    f"{'an ancestor' if operation == 'ANY' else 'a descendant'} "
                    f"table to be specified (none supplied)."
                )
            operand_table = operands[0]._table
            if operation == "ANY":
                if not new_table.is_ancestor(operand_table):
                    raise OperationError(
                        f"'ANY' requires an ancestor table to be specified."
                    )
            elif operation == "THE":
                if not new_table.is_descendant(operand_table):
                    raise OperationError(
                        f"'THE' requires a descendant table to be specified."
                    )
            return new_table

        if operation in ["AND", "OR", "NOT"]:
            operand_table_names = [op._table.name for op in operands]
            if len(set(operand_table_names)) != 1:
                raise OperationError(
                    f"'{operation}' requires all clauses to have the same "
                    f"resolve table."
                )
            return operands[0]._table


class CriteriaClause(Clause):
    @property
    def variable_name(self):
        return self.variable.name


class SelectorClause(CriteriaClause):
    def __init__(self, variable, values, *, include=True, label=None, session=None):
        self.table = variable.table
        self.variable = variable
        self.values = values

        self.include = include
        self.label = label
        self.session = session

    def _to_model_clause(self):
        return aa.Clause(
            criteria=aa.Criteria(
                variable_name=self.variable_name,
                include=self.include,
                logic="OR",
                ignore_case=False,
                text_match_type="Is",
                value_rules=[
                    aa.ValueRule(
                        list_rule=aa.ListRule(
                            list="\t".join(self.values),
                            variable_name=self.variable_name,
                        )
                    )
                ],
                table_name=self.table_name,
                name=self.label,
            )
        )


class CombinedCategoriesClause(CriteriaClause):
    def __init__(self, variable, value_sets, *, include=True, label=None, session=None):
        self.table = variable.table
        self.variable = variable
        self.value_sets = value_sets

        self.include = include
        self.label = label
        self.session = session

    def _to_model_clause(self):
        return aa.Clause(
            criteria=aa.Criteria(
                variable_name=self.variable_name,
                include=self.include,
                logic="OR",
                ignore_case=False,
                text_match_type="Is",
                value_rules=[
                    aa.ValueRule(
                        list_rule=aa.ListRule(
                            list="\t".join(vals), variable_name=var_name
                        )
                    )
                    for var_name, vals in self.value_sets.items()
                ],
                table_name=self.table_name,
                name=self.label,
            )
        )


class NumericClause(CriteriaClause):
    def __init__(self, variable, values, *, include=True, label=None, session=None):
        self.table = variable.table
        self.variable = variable
        self.values = values

        self.include = include
        self.label = label
        self.session = session

    def _to_model_clause(self):
        return aa.Clause(
            criteria=aa.Criteria(
                variable_name=self.variable_name,
                include=self.include,
                logic="OR",
                ignore_case=False,
                text_match_type="Is",
                value_rules=[
                    aa.ValueRule(
                        list_rule=aa.ListRule(
                            list="\t".join(self.values),
                            variable_name=self.variable_name,
                        )
                    )
                ],
                table_name=self.table_name,
                name=self.label,
            )
        )


class TextClause(CriteriaClause):
    def __init__(
        self,
        variable,
        values,
        match_type="Is",
        match_case=True,
        *,
        include=True,
        label=None,
        session=None,
    ):
        self.table = variable.table
        self.variable = variable
        self.values = values
        self.match_type = match_type
        self.match_case = match_case

        self.include = include
        self.label = label
        self.session = session

    def _to_model_clause(self):
        return aa.Clause(
            criteria=aa.Criteria(
                variable_name=self.variable_name,
                include=self.include,
                logic="OR",
                ignore_case=not self.match_case,
                text_match_type=self.match_type,
                value_rules=[
                    aa.ValueRule(
                        list_rule=aa.ListRule(
                            list="\t".join(self.values),
                            variable_name=self.variable_name,
                        )
                    )
                ],
                table_name=self.table_name,
                name=self.label,
            )
        )


class ArrayClause(CriteriaClause):
    def __init__(
        self, variable, values, logic="OR", *, include=True, label=None, session=None
    ):
        self.table = variable.table
        self.variable = variable
        self.values = values
        self.logic = logic

        self.include = include
        self.label = label
        self.session = session

    def _to_model_clause(self):
        return aa.Clause(
            criteria=aa.Criteria(
                variable_name=self.variable_name,
                include=self.include,
                logic=self.logic,
                ignore_case=False,
                text_match_type="Is",
                value_rules=[
                    aa.ValueRule(
                        list_rule=aa.ListRule(
                            list="\t".join(self.values),
                            variable_name=self.variable_name,
                        )
                    )
                ],
                table_name=self.table_name,
                name=self.label,
            )
        )


class FlagArrayClause(ArrayClause):
    pass


class DateListClause(CriteriaClause):
    def __init__(self, variable, values, *, include=True, label=None, session=None):
        self.table = variable.table
        self.variable = variable
        self.values = values

        self.include = include
        self.label = label
        self.session = session

    def _to_model_clause(self):
        return aa.Clause(
            criteria=aa.Criteria(
                variable_name=self.variable_name,
                include=self.include,
                logic="OR",
                ignore_case=False,
                text_match_type="Is",
                value_rules=[
                    aa.ValueRule(
                        list_rule=aa.ListRule(
                            list="\t".join(self.values),
                            variable_name=self.variable_name,
                        ),
                        predefined_rule="AdhocDates",
                    )
                ],
                table_name=self.table_name,
                name=self.label,
            )
        )


class DateRangeClause(CriteriaClause):
    def __init__(self, variable, start, end, *, include=True, label=None, session=None):
        self.table = variable.table
        self.variable = variable
        self.start = start
        self.end = end

        self.include = include
        self.label = label
        self.session = session

    def _create_range_parameters(self):
        params = {}
        if self.start.lower() == "earliest":
            params["start_range_limit"] = "Earliest"
        else:
            params["start_range_limit"] = "Actual"
            params["range_start_date"] = self.start + "T00:00:00"
        if self.end.lower() == "latest":
            params["end_range_limit"] = "Latest"
        else:
            params["end_range_limit"] = "Actual"
            params["range_end_date"] = self.end + "T00:00:00"
        return params

    def _to_model_clause(self):
        return aa.Clause(
            criteria=aa.Criteria(
                variable_name=self.variable_name,
                include=self.include,
                logic="OR",
                ignore_case=False,
                text_match_type="Is",
                value_rules=[
                    aa.ValueRule(
                        predefined_rule="CustomRule",
                        date_rule=aa.DateRule(
                            pattern_frequency="Daily",
                            pattern_interval=1,
                            pattern_days_of_week=["All"],
                            **self._create_range_parameters(),
                        ),
                    )
                ],
                table_name=self.table_name,
                name=self.label,
            )
        )


class TimeRangeClause(CriteriaClause):
    def __init__(self, variable, start, end, *, include=True, label=None, session=None):
        self.table = variable.table
        self.variable = variable
        self.start = start
        self.end = end

        self.include = include
        self.label = label
        self.session = session

    def _to_model_clause(self):
        return aa.Clause(
            criteria=aa.Criteria(
                variable_name=self.variable_name,
                include=self.include,
                logic="OR",
                ignore_case=False,
                text_match_type="Is",
                value_rules=[
                    aa.ValueRule(
                        predefined_rule="CustomTimeRule",
                        time_rule=aa.TimeRule(
                            range_low=self.start, range_high=self.end
                        ),
                    )
                ],
                table_name=self.table_name,
                name=self.label,
            )
        )


class DateTimeRangeClause(DateRangeClause):
    def _create_range_parameters(self):
        params = {}
        if self.start.lower() == "earliest":
            params["start_range_limit"] = "Earliest"
        else:
            params["start_range_limit"] = "Actual"
            params["range_start_date"] = self.start
        if self.end.lower() == "latest":
            params["end_range_limit"] = "Latest"
        else:
            params["end_range_limit"] = "Actual"
            params["range_end_date"] = self.end
        return params


# TODO: write implementation
class ReferenceClause(CriteriaClause):
    def __init__(self, variable, values, *, include=True, label=None, session=None):
        raise NotImplementedError

    def _to_model_clause(self):
        raise NotImplementedError


class BooleanClause(Clause):
    def __init__(self, table, operation, operands, *, label=None, session=None):
        self.table = table
        self.operation = operation
        self.operands = operands

        self.label = label
        self.session = session

    def _to_model_clause(self):
        return aa.Clause(
            logic=aa.Logic(
                operation=self.operation,
                operands=[op._to_model_clause() for op in self.operands],
                table_name=self.table_name,
                name=self.label,
            )
        )


class TableClause(Clause):
    def __init__(self, table, operation, operand, *, label=None, session=None):
        self.table = table
        self.operation = operation
        self.operand = operand

        self.label = label
        self.session = session

    def _to_model_clause(self):
        return aa.Clause(
            logic=aa.Logic(
                operation=self.operation,
                operands=[self.operand._to_model_clause()],
                table_name=self.table_name,
                name=self.label,
            )
        )


class SubSelectionClause(Clause):
    def __init__(self, selection, *, label=None, session=None):
        self.selection = selection

        self.label = label
        self.session = session

    def _to_model_clause(self):
        return aa.Clause(
            # TODO: this may need to be changed depending on
            #  the final shape of the base py-apteco Selection object
            sub_selection=self.selection._to_model_selection()
        )


class BaseLimitClause(Clause):
    def _to_model_clause(self):
        return aa.Clause(
            sub_selection=aa.SubSelection(selection=self._to_model_selection())
        )


class LimitClause(BaseLimitClause):
    def __init__(
        self,
        clause,
        total=None,
        percent=None,
        fraction=None,
        sample_type="First",
        skip_first=0,
        *,
        label=None,
        session=None,
    ):

        input_flag = tuple((i is not None) for i in (total, percent, fraction))
        if sum(input_flag) != 1:
            raise ValueError(
                "Must specify exactly one of `total`, `percent` or `fraction`"
            )
        self.kind = {0: "Total", 1: "Percent", 2: "Fraction"}[input_flag.index(True)]

        self.total = None
        self.percent = None
        self.fraction = None
        if self.kind == "Total":
            if not isinstance(total, Integral) or int(total) < 1:
                raise ValueError("`total` must be an integer greater than 0")
            self.total = int(total)
        elif self.kind == "Percent":
            if not isinstance(percent, Real) or not (0 < float(percent) < 100):
                raise ValueError("`percent` must be a number between 0–100 (exclusive)")
            self.percent = float(percent)
        elif self.kind == "Fraction":
            if not isinstance(fraction, Rational) or not (0 < float(fraction) < 1):
                raise ValueError(
                    "`fraction` must be a rational number between 0 and 1 (exclusive)"
                )
            self.fraction = Fraction(fraction.numerator, fraction.denominator)
        else:
            raise ValueError("Limit kind not recognised")

        if sample_type.title() not in ("First", "Stratified", "Random"):
            raise ValueError(f"{sample_type} is not a valid sample type")
        self.sample_type = sample_type.title()

        if not isinstance(skip_first, Integral) or int(skip_first) < 0:
            raise ValueError("`skip_first` must be a non-negative integer")
        self.skip_first = int(skip_first)

        self.clause = clause
        self.table = clause.table

        self.label = label
        self.session = session

    def _to_model_selection(self):
        return aa.Selection(
            rule=aa.Rule(clause=self.clause._to_model_clause()),
            limits=aa.Limits(
                sampling=self.sample_type,
                total=self.total,
                type=self.kind,
                start_at=self.skip_first,
                percent=self.percent,
                fraction=aa.Fraction(self.fraction.numerator, self.fraction.denominator)
                if self.fraction is not None
                else None,
            ),
            table_name=self.table_name,
            name=self.label,
        )


def ensure_single_or_range(
    input_value,
    type_,
    convert,
    number_text,
    param_text,
    lower_bound=None,
    upper_bound=None,
):

    if isinstance(input_value, Iterable) and not isinstance(input_value, str):
        if not (isinstance(input_value, tuple) and len(input_value) == 2):
            raise ValueError(
                f"Invalid range given for `{param_text}`"
                f" - must be a tuple of two values."
            )
        try:
            start, end = input_value
            start = ensure_single(
                start,
                type_,
                convert,
                number_text,
                f"start of range",
                lower_bound,
                upper_bound,
            )
            end = ensure_single(
                end,
                type_,
                convert,
                number_text,
                f"end of range",
                lower_bound,
                upper_bound,
            )
        except ValueError as exc:
            exc_msg = f"Invalid range given for `{param_text}` - {exc.args[0]}"
            raise ValueError(exc_msg)
        if not start < end:
            raise ValueError(
                f"Invalid range given for `{param_text}`"
                f" - start of range must be less than the end."
            )
        return "range", (start, end)

    return (
        "single",
        ensure_single(
            input_value,
            type_,
            convert,
            number_text,
            f"`{param_text}`",
            lower_bound,
            upper_bound,
        ),
    )


def ensure_single(
    value, type_, convert, number_text, param_text, lower_bound=None, upper_bound=None
):
    if not isinstance(value, type_):
        raise ValueError(f"{param_text} must be {number_text}")
    value = convert(value)
    if lower_bound is not None and not lower_bound < value:
        raise ValueError(f"{param_text} must be greater than {lower_bound}")
    if upper_bound is not None and not value < upper_bound:
        raise ValueError(f"{param_text} must be less than {upper_bound}")
    return value


class TopNClause(BaseLimitClause):
    def __init__(
        self,
        clause,
        total=None,
        percent=None,
        by=None,
        ascending=False,
        *,
        label=None,
        session=None,
    ):
        self.kind, self.total, self.percent = self._check_numerical_inputs(
            percent, total
        )

        if by is None or not hasattr(by, "name"):
            raise ValueError("`by` must be an ordered variable")
        self.by = by

        if not isinstance(ascending, bool):
            raise ValueError("`ascending` must be a boolean (True or False)")
        self.ascending = ascending

        self.clause = clause
        self.table = clause.table

        self.label = label
        self.session = session

    @staticmethod
    def _check_numerical_inputs(percent, total):
        if total is not None:
            if percent is not None:
                raise ValueError(
                    "Must specify either `total` or `percent`, but not both"
                )
            try:
                single_or_range, total = ensure_single_or_range(
                    total, Integral, int, "an integer", "total", lower_bound=0
                )
                kind = (single_or_range, "total")
                return kind, total, None
            except ValueError:
                raise ValueError(
                    "`total` must be an integer or a tuple of two integers (to indicate a range)"
                ) from None
        elif percent is not None:
            try:
                single_or_range, percent = ensure_single_or_range(
                    percent,
                    Real,
                    float,
                    "a percentage",
                    "percent",
                    lower_bound=0,
                    upper_bound=100,
                )
                kind = (single_or_range, "percent")
                return kind, None, percent
            except ValueError:
                raise ValueError(
                    "`percent` must be a number or a tuple of two numbers (to indicate a range)"
                )
        else:
            raise ValueError("Must specify one of `total` or `percent`")

    def _to_model_selection(self):
        if self.kind[0] == "single":
            if self.kind[1] == "total":
                value = self.total
                percent = "NaN"
            elif self.kind[1] == "percent":
                value = 0
                percent = self.percent
            else:
                raise ValueError(f"Invalid kind: {self.kind}")
            direction = "Bottom" if self.ascending else "Top"
            min_value = "NaN"
            max_value = "NaN"
        elif self.kind[0] == "range":
            if self.kind[1] == "total":
                direction = "Range"
                min_value, max_value = self.total
            elif self.kind[1] == "percent":
                direction = "PercentRange"
                min_value, max_value = self.percent
            else:
                raise ValueError(f"Invalid kind: {self.kind}")
            direction += "BottomUp" if self.ascending else "TopDown"
            value = 0
            percent = "NaN"
        else:
            raise ValueError(f"Invalid kind: {self.kind}")

        return aa.Selection(
            rule=aa.Rule(clause=self.clause._to_model_clause()),
            top_n=aa.TopN(
                variable_name=self.by.name,
                direction=direction,
                value=value,
                percent=percent,
                min_value=min_value,
                max_value=max_value,
            ),
            table_name=self.table_name,
            name=self.label,
        )


class NPerVariableClause(BaseLimitClause):
    def __init__(
        self, clause, n, per, by=None, ascending=False, *, label=None, session=None
    ):
        if not isinstance(n, Integral) or int(n) < 1:
            raise ValueError("`n` must be an integer greater than 0")
        self.n = int(n)

        if per is None or not hasattr(per, "is_selectable"):  # only vars have this attr
            raise ValueError("`per` must be a variable")
        if per.type in (VariableType.ARRAY, VariableType.FLAG_ARRAY):
            raise ValueError("`per` cannot be an Array or Flag Array variable")
        self.per = per

        if by is not None and not hasattr(by, "is_selectable"):  # only vars have attr
            raise ValueError("`by` must be an ordered variable")
        self.by = by

        if not isinstance(ascending, bool):
            raise ValueError("`ascending` must be a boolean (True or False)")
        self.ascending = ascending

        self.clause = clause
        self.table = clause.table

        self.label = label
        self.session = session

    def _to_model_selection(self):
        if self.by is not None:
            grouping_sequence_variable_name = self.by.name
            grouping_ascending = self.ascending
        else:
            grouping_sequence_variable_name = None
            grouping_ascending = None
        return aa.Selection(
            rule=aa.Rule(clause=self.clause._to_model_clause()),
            top_n=aa.TopN(
                grouping_variable_name=self.per.name,
                grouping_sequence_variable_name=grouping_sequence_variable_name,
                grouping_ascending=grouping_ascending,
                group_max=self.n,
            ),
            table_name=self.table_name,
            name=self.label,
        )


class NPerTableClause(BaseLimitClause):
    def __init__(
        self, clause, n, per, by=None, ascending=False, *, label=None, session=None
    ):
        if not isinstance(n, Integral) or int(n) < 1:
            raise ValueError("`n` must be an integer greater than 0")
        self.n = int(n)

        if per is None or not hasattr(per, "is_people"):  # only tables have this attr
            raise ValueError("`per` must be a table")
        self.per = per

        if by is not None and not hasattr(by, "is_selectable"):  # only vars have attr
            raise ValueError("`by` must be an ordered variable")
        self.by = by

        if not isinstance(ascending, bool):
            raise ValueError("`ascending` must be a boolean (True or False)")
        self.ascending = ascending

        self.clause = clause
        self.table = clause.table

        self.label = label
        self.session = session

    def _to_model_selection(self):
        if self.by is not None:
            variable_name = self.by.name
            direction = "First" if self.ascending else "Last"
        else:
            variable_name = None
            direction = None
        return aa.Selection(
            rule=aa.Rule(clause=self.clause._to_model_clause()),
            n_per=aa.NPer(
                recency=aa.RFVRecency(
                    variable_name=variable_name, direction=direction, value=self.n
                ),
                grouping_table_name=self.per.name,
                transactional_table_name=self.table_name,
            ),
            table_name=self.table_name,
            name=self.label,
        )



import apteco_api as aa
import numpy as np
import pandas as pd

from apteco.common import VariableType


class DataGrid:
    def __init__(
        self, columns, selection=None, table=None, max_rows=1000, *, session=None
    ):
        self.columns = columns
        self.selection = selection
        self.table = table
        self.max_rows = max_rows
        self.session = session
        self._check_inputs()
        self._data = self._get_data()

    def to_df(self):
        df = pd.DataFrame(self._data, columns=[v.description for v in self.columns])
        for i, v in enumerate(self.columns):
            df.iloc[:, i] = self._convert_column(df.iloc[:, i], v.type)
        return df

    @staticmethod
    def _convert_column(data: pd.Series, column_type):
        if column_type in (VariableType.SELECTOR, VariableType.TEXT, VariableType.REFERENCE):
            return data.astype(str)
        elif column_type == VariableType.NUMERIC:
            return pd.to_numeric(data)
        elif column_type == VariableType.DATE:
            return pd.to_datetime(data, format="%d-%m-%Y", errors="coerce").dt.date
        elif column_type == VariableType.DATETIME:
            return pd.to_datetime(data, format="%d-%m-%Y %H:%M:%S", errors="coerce")
        else:
            raise ValueError(f"Unrecognised column type: {column_type}")

    def _check_inputs(self):
        if self.session is None:
            raise ValueError("You must provide a valid session (none was given).")
        try:
            self.max_rows = int(self.max_rows)
            assert self.max_rows > 0
        except (ValueError, TypeError, AssertionError) as exc:
            raise ValueError("max_rows must be a number greater than 0") from exc
        if not self.columns:
            raise ValueError(
                "You must specify at least one variable"
                " to use as a column on the data grid (none was given)."
            )
        if self.table is None:
            if self.selection is not None:
                self.table = self.selection.table
            else:
                raise ValueError(
                    "You must specify the resolve table of the data grid"
                    " (none was given)."
                )
        self._check_columns()

    def _check_columns(self):
        for column in self.columns:
            if not column.table.is_ancestor(self.table, allow_same=True):
                raise ValueError(
                    f"The resolve table of the data grid is '{self.table.name}',"
                    f" but the variable '{column.name}' belongs to the"
                    f" '{column.table.name}' table."
                    f"\nOnly variables from the same table as the data grid"
                    f" or from ancestor tables can be used as data grid columns."
                )
            if column.type in (VariableType.ARRAY, VariableType.FLAG_ARRAY):
                raise ValueError(
                    f"The variable '{column.name}' has type '{column.type}'."
                    f"\nArray and Flag Array variables"
                    f" are not currently supported as data grid columns."
                )

    def _create_columns(self):
        return [
            aa.Column(id=str(i), variable_name=v.name, column_header=v.description)
            for i, v in enumerate(self.columns)
        ]

    def _get_export(self):
        export = aa.Export(
            base_query=aa.Query(
                selection=self.selection._to_model_selection()
                if self.selection is not None
                else aa.Selection(table_name=self.table.name)
            ),
            resolve_table_name=self.table.name,
            maximum_number_of_rows_to_browse=self.max_rows,
            return_browse_rows=True,
            columns=self._create_columns(),
        )
        exports_controller = aa.ExportsApi(self.session.api_client)
        export_result = exports_controller.exports_perform_export_synchronously(
            self.session.data_view, self.session.system, export=export
        )
        return export_result

    def _get_data(self):
        export_result = self._get_export()
        return [tuple(row.descriptions.split("\t")) for row in export_result.rows]



from typing import Iterable, List, Optional

import apteco_api as aa

from apteco.cube import Cube
from apteco.datagrid import DataGrid
from apteco.exceptions import get_deprecated_attr
from apteco.query import NPerTableClause, TableMixin
from apteco.variables import VariablesAccessor


class Table(TableMixin):
    """Class representing a FastStats system table."""

    def __init__(
        self,
        name: str,
        singular: str,
        plural: str,
        is_default: bool,
        is_people: bool,
        total_records: int,
        child_relationship: str,
        parent_relationship: str,
        has_children: bool,
        parent_name: str,
        parent: Optional["Table"],
        children: List["Table"],
        ancestors: List["Table"],
        descendants: List["Table"],
        variables: VariablesAccessor,
        *,
        session: Optional["Session"] = None,
    ):
        """

        Args:
            name (str): table reference name
            singular (str): noun for a single item from this table
            plural (str): noun for multiple items from this table
            is_default (bool): whether this is the default table
                for this FastStats system
            is_people (bool): whether this is the table
                representing people in this FastStats system
            total_records (int): total number of records on this table
            child_relationship (str): phrase to relate
                to this table from its parent,
                e.g. 'customer <purchased the> product'
            parent_relationship (str): phrase to relate this table
                to its parent,
                e.g. 'product <was purchased by the> customer'
            has_children (bool): whether this table has any child tables
            parent_name (str): name of this table's parent table
                (an empty string for the master table)
            parent (Table): the parent table of this table
                (``None`` for the master table)
            children (List[Table]): list of child tables of this table
                (an empty list if table has no children)
            ancestors (List[Table]): list of ancestor tables
                of this table (an empty list for the master table)
            descendants (List[Table]): list of descendant tables
                of this table (an empty list if table has no children)
            variables (VariablesAccessor): variables on this table
            session (Session): API session the tables data belongs to

        """
        self.name = name
        self.singular = singular
        self.plural = plural
        self.is_default = is_default
        self.is_people = is_people
        self.total_records = total_records
        self.child_relationship = child_relationship
        self.parent_relationship = parent_relationship
        self.has_children = has_children
        self.parent_name = parent_name
        self.parent = parent
        self.children = children
        self.ancestors = ancestors
        self.descendants = descendants
        self.variables = variables
        self.session = session

        self.table = self
        self._name = self.name

    def is_same(self, other: "Table"):
        """Return whether this table is the same as ``other``.

        Args:
            other (Table): the table to check against

        Returns:
            bool: ``True`` if this is the same as ``other``, otherwise ``False``

        """
        return self.name == other.name

    def is_ancestor(self, other: "Table", allow_same: bool = False):
        """Return whether this table is an ancestor of ``other``.

        Args:
            other (Table): the table to check against
            allow_same (bool): whether to include the table itself as an 'ancestor'
                (default is ``False``)

        Returns:
            bool: ``True`` if this is an ancestor of ``other``, otherwise ``False``

        """
        return self in other.ancestors or (allow_same and self.is_same(other))

    def is_descendant(self, other: "Table", allow_same: bool = False):
        """Return whether this table is a descendant of ``other``.

        Args:
            other (Table): the table to check against
            allow_same (bool): whether to include the table itself as an 'ancestor'
                (default is ``False``)

        Returns:
            bool: ``True`` if this is a descendant of ``other``, otherwise ``False``

        """
        return self in other.descendants or (allow_same and self.is_same(other))

    def is_related(self, other: "Table", allow_same: bool = False):
        """Return whether this table is related to ``other``.

        'related' is defined as being either an ancestor or descendant.

        Args:
            other (Table): the table to check against
            allow_same (bool): whether to include the table itself as 'related'
                (default is ``False``)

        Returns:
            bool: ``True`` if this is related to ``other``, otherwise ``False``

        """
        return (
            self.is_ancestor(other)
            or self.is_descendant(other)
            or (allow_same and self.is_same(other))
        )

    def __eq__(self, other):
        """Return whether this is the same table as ``other``."""
        return self.is_same(other)

    def __lt__(self, other):
        """Return whether this table is an ancestor of ``other``."""
        return self.is_ancestor(other)

    def __gt__(self, other):
        """Return whether this table is a descendant of ``other``."""
        return self.is_descendant(other)

    def __le__(self, other):
        """Return whether this table is the same or an ancestor of ``other``."""
        return self.is_ancestor(other, allow_same=True)

    def __ge__(self, other):
        """Return whether this table is the same or a descendant of ``other``."""
        return self.is_descendant(other, allow_same=True)

    def __getitem__(self, item):
        return self.variables[item]

    def datagrid(self, columns, selection=None, max_rows=1000):
        return DataGrid(
            columns,
            selection=selection,
            table=self,
            max_rows=max_rows,
            session=self.session,
        )

    def cube(self, dimensions, measures=None, selection=None):
        return Cube(
            dimensions, measures, selection=selection, table=self, session=self.session
        )

    def _as_nper_clause(self, clause, n, by, ascending, label):
        return NPerTableClause(
            clause=clause,
            n=n,
            per=self,
            by=by,
            ascending=ascending,
            label=label,
            session=self.session,
        )

    def _to_model_measure(self, table):
        return aa.Measure(
            id=f"{self.plural.title()}",
            resolve_table_name=self.name,
            function="Count",
            variable_name=None,
        )

    def __getattr__(self, item):
        DEPRECATED_ATTRS = {
            "singular_display_name": ("singular", "0.6.0"),
            "plural_display_name": ("plural", "0.6.0"),
            "is_default_table": ("is_default", "0.6.0"),
            "is_people_table": ("is_people", "0.6.0"),
            "child_relationship_name": ("child_relationship", "0.6.0"),
            "parent_relationship_name": ("parent_relationship", "0.6.0"),
            "has_child_tables": ("has_children", "0.6.0"),
        }

        if item in DEPRECATED_ATTRS:
            return get_deprecated_attr(self, item, *DEPRECATED_ATTRS[item])
        else:
            raise AttributeError(
                f"'{self.__class__.__name__}' has no attribute '{item}'"
            )


class TablesAccessor:
    """List- and dictionary-like access for tables."""

    def __init__(self, tables: Iterable[Table]):
        self._tables = list(tables)
        self._tables_by_name = {table.name: table for table in self._tables}

    def __len__(self):
        return len(self._tables)

    def __iter__(self):
        return iter(self._tables)

    def __getitem__(self, item):
        try:
            return self._tables_by_name[item]
        except KeyError as exc:
            raise KeyError(f"Lookup key '{item}' did not match a table name.") from exc



import apteco_api as aa

__all__ = [
    "Sum",
    "Mean",
    "Populated",
    "Min",
    "Max",
    "Median",
    "Mode",
    "Variance",
    "StdDev",
    "LowerQuartile",
    "UpperQuartile",
    "InterQuartileRange",
    "CountDistinct",
    "CountMode",
]

from apteco.common import VariableType


def _ensure_correct_type(operand, accepted_types):
    try:
        assert operand.type in accepted_types
    except (AssertionError, AttributeError):
        raise ValueError(
            f"The operand must be a {' or '.join(t for t in accepted_types)} variable"
        ) from None


class Statistic:
    def __init__(self, operand, *, label=None):
        self.table = operand.table
        _ensure_correct_type(operand, self._accepted_types)
        self.operand = operand

        self.label = label

        self._name = f"{self._display_name}({self.operand.description})"

    def _to_model_measure(self, table):
        return aa.Measure(
            id=self.label if self.label is not None else self._name,
            resolve_table_name=table.name,
            function=self._model_function,
            variable_name=self.operand.name,
        )


class Sum(Statistic):
    _model_function = "Sum"
    _display_name = "Sum"
    _accepted_types = (VariableType.NUMERIC,)


class Mean(Statistic):
    _model_function = "Mean"
    _display_name = "Mean"
    _accepted_types = (VariableType.NUMERIC,)


class Populated(Statistic):
    _model_function = "VariableCount"
    _display_name = "Populated"
    _accepted_types = (VariableType.NUMERIC,)


class Min(Statistic):
    _model_function = "Minimum"
    _display_name = "Min"
    _accepted_types = (VariableType.NUMERIC,)


class Max(Statistic):
    _model_function = "Maximum"
    _display_name = "Max"
    _accepted_types = (VariableType.NUMERIC,)


class Median(Statistic):
    _model_function = "Median"
    _display_name = "Median"
    _accepted_types = (VariableType.NUMERIC,)


class Mode(Statistic):
    _model_function = "Mode"
    _display_name = "Mode"
    _accepted_types = (VariableType.NUMERIC,)


class Variance(Statistic):
    _model_function = "Variance"
    _display_name = "Variance"
    _accepted_types = (VariableType.NUMERIC,)


class StdDev(Statistic):
    _model_function = "StandardDeviation"
    _display_name = "Std Dev"
    _accepted_types = (VariableType.NUMERIC,)


class LowerQuartile(Statistic):
    _model_function = "LowerQuartile"
    _display_name = "Lower Quartile"
    _accepted_types = (VariableType.NUMERIC,)


class UpperQuartile(Statistic):
    _model_function = "UpperQuartile"
    _display_name = "Upper Quartile"
    _accepted_types = (VariableType.NUMERIC,)


class InterQuartileRange(Statistic):
    _model_function = "InterQuartileRange"
    _display_name = "Inter Quartile Range"
    _accepted_types = (VariableType.NUMERIC,)


class CountDistinct(Statistic):
    _model_function = "CountDistinct"
    _display_name = "Count Distinct"
    _accepted_types = (VariableType.NUMERIC, VariableType.SELECTOR)


class CountMode(Statistic):
    _model_function = "MaxDistinctCount"
    _display_name = "Count Mode"
    _accepted_types = (VariableType.NUMERIC, VariableType.SELECTOR)



import getpass
import json
import logging
import warnings
from collections import Counter, defaultdict, namedtuple
from json import JSONDecodeError
from typing import Any, Dict, List, Tuple

import apteco_api as aa
import PySimpleGUI

from apteco.data.apteco_logo import APTECO_LOGO
from apteco.exceptions import (
    ApiResultsError,
    DeserializeError,
    TablesError,
    VariablesError,
)
from apteco.tables import Table, TablesAccessor
from apteco.variables import (
    ArrayVariable,
    DateTimeVariable,
    DateVariable,
    FlagArrayVariable,
    NumericVariable,
    ReferenceVariable,
    SelectorVariable,
    TextVariable,
    Variable,
    VariablesAccessor,
)

NOT_ASSIGNED: Any = object()
VARIABLES_PER_PAGE = 1000


class Session:
    def __init__(self, credentials: "Credentials", system: str):
        self._unpack_credentials(credentials)
        self._create_client()
        self.system = system
        self._fetch_system_info()
        tables_without_vars, master_table_name = InitializeTablesAlgorithm(self).run()
        variables, tables = InitializeVariablesAlgorithm(
            self, tables_without_vars
        ).run()
        self.tables = TablesAccessor(tables)
        self.variables = VariablesAccessor(variables)
        self.master_table = self.tables[master_table_name]

    def _unpack_credentials(self, credentials):
        """Copy credentials data into session."""
        self.base_url = credentials.base_url
        self.data_view = credentials.data_view
        self.session_id = credentials.session_id
        self.access_token = credentials.access_token
        self.user = credentials.user

    def _create_client(self):
        """Create an authorized API client."""
        config = aa.Configuration()
        config.host = self.base_url
        config.api_key = {"Authorization": self.access_token}
        config.api_key_prefix = {"Authorization": "Bearer"}
        self._config = config
        self.api_client = aa.ApiClient(configuration=self._config)

    def _fetch_system_info(self):
        """Fetch FastStats system info from API and add to session."""
        systems_controller = aa.FastStatsSystemsApi(self.api_client)
        result = systems_controller.fast_stats_systems_get_fast_stats_system(
            self.data_view, self.system
        )
        self.system_info = FastStatsSystem(
            name=result.name,
            description=result.description,
            build_date=result.fast_stats_build_date,
            view_name=result.view_name,
        )

    def _to_dict(self):
        return {
            "base_url": self.base_url,
            "data_view": self.data_view,
            "session_id": self.session_id,
            "access_token": self.access_token,
            "user": self.user._asdict(),
            "system": self.system,
        }

    @staticmethod
    def _from_dict(d):
        try:
            credentials = Credentials(
                d["base_url"],
                d["data_view"],
                d["session_id"],
                d["access_token"],
                User(**d["user"]),
            )
            system = d["system"]
        except KeyError as e:
            raise DeserializeError(f"Data missing from 'Session' object: no {e} found.")
        except TypeError as exc:  # arguments missing from User __new__() call
            raise DeserializeError(
                f"The following parameter(s) were missing from 'User' object: "
                f"{exc.args[0].split(':')[1].strip()}"
            )
        else:
            return Session(credentials, system)

    def serialize(self):
        return json.dumps(self._to_dict())

    @staticmethod
    def deserialize(s: str):
        try:
            d = json.loads(s)
        except JSONDecodeError:
            raise DeserializeError("The given input could not be deserialized.")
        else:
            return Session._from_dict(d)


User = namedtuple("User", ["username", "first_name", "surname", "email_address"])


class Credentials:
    """Class to hold credentials from the simple login process."""

    def __init__(
        self,
        base_url: str,
        data_view: str,
        session_id: str,
        access_token: str,
        user: User,
    ):
        """

        Args:
            base_url (str): API base URL, normally ending '/OrbitAPI'
            data_view (str): name of data view
            session_id (str): Apteco session ID
            access_token (str): access token for current session
            user (User): API user

        """
        self.base_url = base_url
        self.data_view = data_view
        self.session_id = session_id
        self.access_token = access_token
        self.user = user


FastStatsSystem = namedtuple(
    "FastStatsSystem", ["name", "description", "build_date", "view_name"]
)


def login(base_url: str, data_view: str, system: str, user: str) -> Session:
    """Log in to the API without supplying password directly.

    Args:
        base_url (str): API base URL, normally ending '/OrbitAPI'
        data_view (str): DataView being logged into
        system (str): FastStats system to connect to
        user (str): username of API user

    Returns:
        Session: API session object

    """
    return login_with_password(
        base_url, data_view, system, user, password=_get_password()
    )


def login_with_password(
    base_url: str, data_view: str, system: str, user: str, password: str
) -> Session:
    """Log in to the API, supplying password directly.

    Args:
        base_url (str): API base URL, normally ending '/OrbitAPI'
        data_view (str): DataView being logged into
        system (str): FastStats system to connect to
        user (str): username of API user
        password (str): password for this user

    Returns:
        Session: API session object

    """
    credentials = SimpleLoginAlgorithm(base_url, data_view).run(user, password)
    return Session(credentials, system)


def _get_password(prompt: str = "Enter your password: ") -> str:
    """Get password from user without displaying it on screen."""
    try:
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            return getpass.getpass(prompt)
    except getpass.GetPassWarning:
        return PySimpleGUI.PopupGetText(
            prompt,
            password_char="*",
            title="Apteco API",
            button_color=("#ffffff", "#004964"),
            icon=APTECO_LOGO,
        )


# https://gieseanw.wordpress.com/2019/05/10/algorithms-as-objects/
class SimpleLoginAlgorithm:
    """Class holding the algorithm to carry out a simple login.

    Attributes:
        base_url (str): API base URL, normally ending '/OrbitAPI'
        data_view (str): DataView being logged into
        api_client (aa.ApiClient): API client used to log in
        session_id (str): Apteco session ID for the created session
        access_token (str): access token for the created session
        user (User): API user

    Methods:
        run(): entry point to run the algorithm

    """

    def __init__(self, base_url: str, data_view: str):
        """

        Args:
            base_url (str): API base URL, normally ending '/OrbitAPI'
            data_view (str): DataView being logged into

        """
        self.base_url = base_url
        self.data_view = data_view

    def run(self, user: str, password: str) -> Credentials:
        """Run the algorithm with the given login credentials.

        Args:
            user (str): username of API user
            password (str): password for this user

        Returns:
            Credentials: API session credentials

        """
        self._create_unauthorized_client()
        self._simple_login(user, password)
        self._create_credentials()
        return self.credentials

    def _create_unauthorized_client(self):
        """Create an unauthorized API client."""
        config = aa.Configuration()
        config.host = self.base_url
        self._config = config
        self.api_client = aa.ApiClient(configuration=self._config)

    def _simple_login(self, user, password):
        """Call API to perform simple login."""
        sessions_controller = aa.SessionsApi(self.api_client)
        login_response = sessions_controller.sessions_create_session_simple(
            self.data_view, user, password
        )
        self.session_id = login_response.session_id
        self.access_token = login_response.access_token
        self.user = User(
            username=login_response.user.username,
            first_name=login_response.user.firstname,
            surname=login_response.user.surname,
            email_address=login_response.user.email_address,
        )

    def _create_credentials(self):
        """Initialize session credentials object."""
        self.credentials = Credentials(
            self.base_url, self.data_view, self.session_id, self.access_token, self.user
        )


class InitializeTablesAlgorithm:
    """Class holding the algorithm to initialize system tables.

    The purpose of this algorithm is to
    retrieve the raw tables data for the given system,
    process it to infer parent and child table relationships,
    convert the raw tables into py-apteco ``Table`` objects,
    and assign to these tables their relationships to other tables.

    Attributes:
        data_view (str): DataView the system belongs to
        system (str): FastStats system the session is connected to
        api_client (aa.ApiClient): client to handle API calls
        session (Session): API session the tables data belongs to
        raw_tables (List[aa.Table]): list of raw tables
        children_lookup (Dict[str, List[str]]):
            mapping from table name to list of its child table names
        master_table (Table): master table of the FastStats system
        tables_lookup (Dict[str, Table]):
            mapping from table name to its ``Table`` object

    Methods:
        run(): entry point to run the algorithm

    """

    def __init__(self, session):
        """

        Args:
            session (Session): API session the tables data belongs to

        """
        self.data_view = session.data_view
        self.system = session.system
        self.api_client = session.api_client
        self.session = session

    def run(self) -> Tuple[Dict[str, Table], str]:
        """Run the algorithm.

        Returns:
            (tuple): tuple containing:

                tables_lookup (Dict[str, Table]):
                    mapping from table name to its ``Table`` object
                master_table_name (str):
                    name of the master table of the FastStats system

        """
        self._get_raw_tables()
        self._identify_children()
        self._create_tables()
        self._assign_parent_and_children()
        self._find_master_table()
        _tree_tables = self._assign_ancestors_and_descendants(self.master_table, [])
        self._check_all_tables_in_tree(_tree_tables)
        self._check_all_relations_assigned()
        return self.tables_lookup, self.master_table.name

    def _get_raw_tables(self):
        """Get list of all tables from API."""
        systems_controller = aa.FastStatsSystemsApi(self.api_client)
        results = systems_controller.fast_stats_systems_get_fast_stats_tables(
            self.data_view, self.system, count=1000
        )
        self._check_table_results_consistency(results)
        self.raw_tables = results.list

    @staticmethod
    def _check_table_results_consistency(results: aa.PagedResultsTable):
        """Check the number of tables in list matches stated count."""
        results_count = results.count
        list_count = len(results.list)
        if not results_count == list_count:
            raise ApiResultsError(
                f"API stated {results_count} tables were returned"
                f" but {list_count} were found."
            )

    def _identify_children(self):
        """Identify child tables for each table."""
        self.children_lookup = defaultdict(list)
        for table in self.raw_tables:
            self.children_lookup[table.parent_table].append(table.name)
        # don't freeze yet: will need to look up childless tables and return empty list

    def _create_tables(self):
        """Create py-apteco tables from apteco-api ones."""
        self.tables_lookup = {
            t.name: Table(
                t.name,
                t.singular_display_name,
                t.plural_display_name,
                t.is_default_table,
                t.is_people_table,
                t.total_records,
                t.child_relationship_name,
                t.parent_relationship_name,
                t.has_child_tables,
                t.parent_table,
                NOT_ASSIGNED,
                NOT_ASSIGNED,
                NOT_ASSIGNED,
                NOT_ASSIGNED,
                NOT_ASSIGNED,
                session=self.session,
            )
            for t in self.raw_tables
        }

    def _assign_parent_and_children(self):
        """Assign parent and children attributes for each table."""
        for table in self.tables_lookup.values():
            if table.parent_name == "":
                table.parent = None
            else:
                table.parent = self.tables_lookup[table.parent_name]
            table.children = [
                self.tables_lookup[name] for name in self.children_lookup[table.name]
            ]
            self._check_child_tables_consistency(table)
        self.children_lookup.default_factory = None  # 'freeze' as normal dict

    @staticmethod
    def _check_child_tables_consistency(table: Table):
        """Check table's children matches ``has_children``."""
        if table.has_children and not table.children:
            raise TablesError(
                f"API stated '{table.name}' table has child tables but none were found."
            )
        if not table.has_children and table.children:
            raise TablesError(
                f"API stated '{table.name}' table has no child tables"
                f" but {len(table.children)} were found."
            )

    def _find_master_table(self):
        """Retrieve master table, ensuring there is exactly one."""
        try:
            (master_table_name,) = self.children_lookup[""]
        except KeyError:
            raise TablesError("No master table found.")
        except ValueError:  # unpacking failed => !=1 master tables
            raise TablesError(
                f"Found {len(self.children_lookup[''])} master tables,"
                f" there should be 1."
            )
        try:
            self.master_table = self.tables_lookup[master_table_name]
        except KeyError:
            raise TablesError(
                f"The master table '{master_table_name}' could not be found."
            )

    def _assign_ancestors_and_descendants(
        self, table: Table, ancestors: List[Table]
    ) -> List[Table]:
        """Assign ancestor and descendant tables for each table."""
        table.ancestors = ancestors
        table.descendants = [
            descendant
            for child in table.children
            for descendant in self._assign_ancestors_and_descendants(
                child, [table] + table.ancestors
            )
        ]
        return [table] + table.descendants

    def _check_all_tables_in_tree(self, _tree_tables):
        """Check all tables appear in table tree exactly once."""
        tree_tables_counter = Counter(t.name for t in _tree_tables)
        raw_tables_counter = Counter(t.name for t in self.raw_tables)
        if not tree_tables_counter == raw_tables_counter:
            diff = Counter(tree_tables_counter)
            diff.subtract(raw_tables_counter)
            raise TablesError(
                f"Error constructing table tree:"
                f" {len(+diff)} table(s) occurred more than once in tree"
                f" and {len(-diff)} table(s) did not occur at all."
            )

    def _check_all_relations_assigned(self):
        """Check tables have all relation attributes assigned."""
        relations = ["parent", "children", "ancestors", "descendants"]
        no_relation = defaultdict(list)
        for table in self.tables_lookup.values():
            for rel in relations:
                if getattr(table, rel) is NOT_ASSIGNED:
                    no_relation[rel].append(table.name)
        error_details = ""
        for rel in relations:
            if no_relation[rel]:
                error_details += (
                    f"\n{len(no_relation[rel])} table(s) had no {rel} assigned."
                    f" First example: '{no_relation[rel][0]}' table"
                )
        if error_details:
            raise TablesError("Error constructing table tree:" + error_details)


class InitializeVariablesAlgorithm:
    """Class holding the algorithm to initialize system variables.

    The purpose of this algorithm is to
    retrieve the raw variables for the given system,
    convert them into py-apteco ``Variable`` objects,
    and assign these to their tables.

    Attributes:
        data_view (str): DataView the system belongs to
        system (str): FastStats system the session is connected to
        api_client (aa.ApiClient): client to handle API calls
        session (Session): API session the variables data belongs to
        tables_lookup (Dict[str, Table]):
            mapping from table name to its ``Table`` object,
            initially with ``variables`` attribute as ``NOT_ASSIGNED``
        raw_variables (List[aa.Variable]): list of raw variables
        variables (List[Variable]): list of variables
            as py-apteco ``Variable`` objects
        variables_lookup (Dict[str, List[Variable]]):
            mapping from table name to list of its variables

    Methods:
        run(): entry point to run the algorithm

    """

    def __init__(self, session, tables_without_variables):
        """

        Args:
            session (Session): API session the variables data belongs to
            tables_without_variables (Dict[str, Table]):
                mapping from table name to its ``Table`` object,
                with variables attribute as ``NOT_ASSIGNED``

        """
        self.data_view = session.data_view
        self.system = session.system
        self.api_client = session.api_client
        self.session = session
        self.tables_lookup = tables_without_variables

    def run(self) -> Tuple[List[Variable], List[Table]]:
        """Run the algorithm.

        Returns:
            (tuple): tuple containing:

                variables (List[Variable]):
                    list of variables as py-apteco ``Variable`` objects
                tables (List[Table]):
                    list of tables as py-apteco ``Table`` objects

        """

        self._get_raw_variables()
        self._create_variables()
        self._identify_variables()
        self._assign_variables()
        self._check_all_variables_assigned()
        return self.variables, list(self.tables_lookup.values())

    def _get_raw_variables(self, variables_per_page=VARIABLES_PER_PAGE):
        """Get list of all variables from API."""
        systems_controller = aa.FastStatsSystemsApi(self.api_client)

        self.raw_variables = []
        offset = 0
        while True:
            results = systems_controller.fast_stats_systems_get_fast_stats_variables(
                self.data_view, self.system, count=variables_per_page, offset=offset
            )  # type: aa.PagedResultsVariable
            self.raw_variables.extend(results.list)
            if results.offset + results.count >= results.total_count:
                break
            offset = results.offset + results.count
        self._check_variable_results_consistency(results.total_count)

    def _check_variable_results_consistency(self, total_variables: int):
        """Check number of variables returned matches stated total count."""
        list_count = len(self.raw_variables)
        if not total_variables == list_count:
            raise ApiResultsError(
                f"API stated there are {total_variables} variables in the system"
                f" but {list_count} were returned."
            )

    def _create_variables(self):
        """Create py-apteco variables from apteco-api ones."""
        self.variables = []
        for v in self.raw_variables:
            try:
                variable_class = self._choose_variable(v)
            except VariablesError as exc:
                exc_msg = exc.args[0]
                logging.warning(exc_msg)
            else:
                variable = variable_class(
                    name=v.name,
                    description=v.description,
                    type=v.type,
                    folder_name=v.folder_name,
                    table=self.tables_lookup[v.table_name],
                    is_selectable=v.is_selectable,
                    is_browsable=v.is_browsable,
                    is_exportable=v.is_exportable,
                    is_virtual=v.is_virtual,
                    selector_info=v.selector_info,
                    numeric_info=v.numeric_info,
                    text_info=v.text_info,
                    reference_info=v.reference_info,
                    session=self.session,
                )
                self.variables.append(variable)

    @staticmethod
    def _choose_variable(raw_variable: aa.Variable):
        """Get class to create given variable according to its type."""
        variable_type_lookup = {
            ("Selector", "Categorical", "SingleValue", False): SelectorVariable,
            ("Selector", "Categorical", "SingleValue", True): SelectorVariable,  # should be Combined Categories type
            ("Numeric", None, None, False): NumericVariable,
            ("Text", None, None, False): TextVariable,
            ("Selector", "Categorical", "OrArray", False): ArrayVariable,
            ("Selector", "Categorical", "OrArray", True): ArrayVariable,  # should be Combined Categories type
            ("Selector", "Categorical", "OrBitArray", False): FlagArrayVariable,
            ("Selector", "Categorical", "OrBitArray", True): FlagArrayVariable,  # should be Combined Categories type
            ("Selector", "Date", "SingleValue", False): DateVariable,
            ("Selector", "Date", "OrBitArray", False): FlagArrayVariable,  # some kind of VV
            ("Selector", "DateTime", "SingleValue", False): DateTimeVariable,
            ("Reference", None, None, False): ReferenceVariable,
        }

        determinant = (
            raw_variable.type,
            (raw_variable.to_dict().get("selector_info") or {}).get("sub_type"),
            (raw_variable.to_dict().get("selector_info") or {}).get("selector_type"),
            bool(
                (raw_variable.to_dict().get("selector_info") or {}).get(
                    "combined_from_variable_name"
                )
            ),
        )
        try:
            return variable_type_lookup[determinant]
        except KeyError as exc:
            raise VariablesError(
                f"Failed to initialize variable '{raw_variable.name}',"
                f" did not recognise the type from determinant: {determinant}"
            ) from exc

    def _identify_variables(self):
        """Identify variables for each table."""
        self.variables_lookup = defaultdict(list)
        for variable in self.variables:
            self.variables_lookup[variable.table_name].append(variable)
        self.variables_lookup.default_factory = None  # 'freeze' as normal dict

    def _assign_variables(self):
        """Assign variables to each table."""
        for table in self.tables_lookup.values():
            try:
                table.variables = VariablesAccessor(self.variables_lookup[table.name])
            except KeyError:
                raise VariablesError(f"No variables found for '{table.name}' table.")

    def _check_all_variables_assigned(self):
        """Check all tables have variables attribute assigned."""
        no_variables = []
        for table in self.tables_lookup.values():
            if table.variables is NOT_ASSIGNED:
                no_variables.append(table.name)
        if no_variables:
            raise VariablesError(
                f"{len(no_variables)} table(s) had no variables assigned."
            )



__version__ = "0.8.0"

__all__ = ["login", "login_with_password", "Session", "DataGrid", "Cube"]

from .session import login, login_with_password, Session
from .datagrid import DataGrid
from .cube import Cube



from enum import Enum


class VariableType(str, Enum):
    SELECTOR = "Selector"
    # COMBINED_CATEGORIES = "CombinedCategories"
    NUMERIC = "Numeric"
    TEXT = "Text"
    ARRAY = "Array"
    FLAG_ARRAY = "FlagArray"
    DATE = "Date"
    DATETIME = "DateTime"
    BANDED_DATE = "BandedDate"
    REFERENCE = "Reference"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['apteco', 'apteco.data']

package_data = \
{'': ['*']}

install_requires = \
['apteco-api>=0.2.1,<0.3.0',
 'numpy>=1.18,<2.0',
 'pandas>=1.0,<2.0',
 'pysimplegui>=4.0,<5.0']

setup_kwargs = {
    'name': 'apteco',
    'version': '0.8.0',
    'description': 'A Python package for interacting with Apteco Marketing Suite resources via the Apteco API.',
    'long_description': '**************\n  Quickstart\n**************\n\nRequirements\n============\n\n* Python 3.6+\n* Access to an installation of the Apteco API\n\nThe Apteco API (which also goes under the name **Orbit API**)\nis part of the Apteco Orbit™ installation.\nIf you have access to Apteco Orbit™, you also have access to the Apteco API!\nIf you\'re not sure about this, contact whoever administers your Apteco software,\nor get in touch with Apteco support (support@apteco.com).\n\nInstallation\n============\n\nYou can install the package the usual way from PyPI using ``pip``:\n\n.. code-block:: console\n\n   python -m pip install apteco\n\nLogging in\n==========\n\nYour login credentials are the same username and password\nyou would use to log in to Apteco Orbit™:\n\n.. code-block:: python\n\n    >>> from apteco import login\n    >>> my_session = login("https://my-site.com/OrbitAPI", "my_data_view", "my_system", "jdoe")\n\nYou will be asked to enter your password in the terminal, which won\'t be echoed.\nIf Python is unable to ask for your password in this way,\nit will provide a pop-up box instead.\nThis might appear in the background,\nso check your taskbar for a new window if nothing seems to be happening.\n\nIf you don\'t want to enter your password every time,\nthere\'s also a ``login_with_password`` function which takes your password\nas a fifth argument:\n\n.. code-block:: python\n\n    >>> from apteco import login_with_password\n    >>> my_session = login_with_password(\n    ...     "https://my-site.com/OrbitAPI",\n    ...     "my_data_view",\n    ...     "my_system",\n    ...     "jdoe",\n    ...     "password",  # password is in plain sight in the code!\n    ... )\n\nTables\n======\n\nTables are accessed through the ``tables`` attribute on the ``Session`` object.\nYou can retrieve a table using its name:\n\n.. code-block:: python\n\n    >>> bookings = my_session.tables["Bookings"]\n\n``Table`` objects have properties for various metadata:\n\n.. code-block:: python\n\n    >>> print(\n    ...     f"There are {bookings.total_records:,}"\n    ...     f" {bookings.plural.lower()}"\n    ...     f" in the system."\n    ... )\n    There are 2,130,081 bookings in the system.\n\nVariables\n=========\n\nVariables are accessed through the ``variables`` attribute\non the ``Session`` object.\nYou can retrieve a variable using its name or description:\n\n.. code-block:: python\n\n    >>> occupation = my_session.variables["peOccu"]  # name\n    >>> cost = my_session.variables["Cost"]  # description\n\nEach table also has a ``variables`` attribute\nfor accessing the variables on that table:\n\n.. code-block:: python\n\n    >>> occupation = people.variables["peOccu"]\n    >>> cost = bookings.variables["Cost"]\n\nFor convenience you can access variables by indexing into the ``Table`` itself:\n\n.. code-block:: python\n\n    >>> occupation = people["peOccu"]\n    >>> cost = bookings["Cost"]\n\n``Variable`` objects have attributes with various metadata:\n\n.. code-block:: python\n\n    >>> occupation.type\n    <VariableType.SELECTOR: \'Selector\'>\n    >>> cost.description\n    \'Cost\'\n\nCreating selections\n===================\n\nYou can use the Python operators with ``Variable`` objects to build selections\nbased on criteria and return a count:\n\n.. code-block:: python\n\n    >>> sweden = bookings["Destination"] == "29"\n    >>> sweden.count()\n    25207\n\nYou can specify multiple values using any *iterable*:\n\n.. code-block:: python\n\n    >>> people = my_session.tables["People"]\n    >>> high_earners = people["Income"] == (f"{i:02}" for i in range(7, 12))\n    >>> high_earners.count()\n    7114\n\nYou can use other operators as well; for example, to exclude values:\n\n.. code-block:: python\n\n    >>> households = my_session.tables["Households"]\n    >>> uk_only = households["Region"] != "14"  # 14 is Channel Islands\n    >>> uk_only.count()\n    741572\n\nOr to allow a range of values:\n\n.. code-block:: python\n\n    >>> low_profit = bookings["Profit"] <= 25\n    >>> low_profit.count()\n    211328\n\n.. code-block:: python\n\n    >>> second_half_of_alphabet = people["Surname"] >= "N"\n    >>> second_half_of_alphabet.count()\n    410954\n\nDate and DateTime variables use the built-in ``datetime`` module:\n\n.. code-block:: python\n\n    >>> from datetime import date, datetime\n    >>> bookings_before_2019 = bookings["Booking Date"] <= date(2018, 12, 31)\n    >>> bookings_before_2019.count()\n    972439\n\nYou can take advantage of functionality available in other Python packages:\n\n.. code-block:: python\n\n    >>> from dateutil.relativedelta import relativedelta\n    >>> under_30 = people["DOB"] >= date.today() - relativedelta(years=30)\n    >>> under_30.count()\n    207737\n\nCombining selections\n====================\n\nYou can use the ``&`` ``|`` operators to combine selection criteria:\n\n.. code-block:: python\n\n    >>> sweden = bookings["Destination"] == "29"\n    >>> cost_at_least_2k = bookings["Cost"] >= 2000\n    >>> expensive_sweden = sweden & cost_at_least_2k\n    >>> expensive_sweden.count()\n    632\n    >>> student = people["Occupation"] == "4"\n    >>> under_21 = people["DOB"] >= date.today() - relativedelta(years=21)\n    >>> eligible_for_discount = student | under_21\n    >>> eligible_for_discount.count()\n    188364\n\nThe ``~`` operator negates a selection:\n\n.. code-block:: python\n\n    >>> pay_full_price = ~eligible_for_discount\n    >>> pay_full_price.count()\n    968189\n\nYou can join clauses from different tables and it will automatically handle\nthe required table changes:\n\n.. code-block:: python\n\n    >>> high_affordability = high_earners | cost_at_least_2k  # will resolve to people\n    >>> high_affordability.count()\n    56096\n    >>> high_affordability.table_name\n    \'People\'\n\nThe left-most clause determines the resolve table:\n\n.. code-block:: python\n\n    >>> female = people["Gender"] == "F"\n    >>> usa = bookings["Destination"] == "38"\n    >>> female.table_name\n    \'People\'\n    >>> usa.table_name\n    \'Bookings\'\n    >>> (female & usa).table_name\n    \'People\'\n    >>> (usa & female).table_name\n    \'Bookings\'\n\nYou can manually set the resolve table using the ``*`` operator:\n\n.. code-block:: python\n\n    >>> bookings_by_under_21s = bookings * under_21\n    >>> bookings_by_under_21s.count()\n    135100\n    >>> bookings_by_under_21s.table_name\n    \'Bookings\'\n\nCompound clauses follow Python operator precedence:\n\n.. code-block:: python\n\n    >>> student_or_young_female = student | female & under_21\n    >>> student_or_young_female.count()\n    166708\n    >>> student_or_female_must_be_young = (student | female) & under_21\n    >>> student_or_female_must_be_young.count()\n    49225\n\nBe especially careful where compound clauses involve table changes:\n\n.. code-block:: python\n\n    >>> women_to_sweden = female & sweden\n    >>> women_to_sweden.count()  # selection on People table\n    8674\n    >>> audience_1 = bookings * (female & sweden)\n    >>> audience_1.count()  # bookings by women who\'ve been to sweden\n    23553\n    >>> audience_2 = (bookings * female) & sweden\n    >>> audience_2.count()  # bookings made by a woman, with destination of sweden\n    8687\n\nCreating data grids\n===================\n\nYou can create a data grid from a table:\n\n.. code-block:: python\n\n    >>> urn = bookings["Booking URN"]\n    >>> dest = bookings["Destination"]\n    >>> occupation = people["Occupation"]\n    >>> town = households["Town"]\n    >>> datagrid = bookings.datagrid([urn, dest, cost, occupation, town])\n\nConvert it to a Pandas DataFrame:\n\n.. code-block:: python\n\n    >>> datagrid.to_df()\n        Booking URN    Destination     Cost       Occupation        Town\n    0      10001265         France  1392.35  Sales Executive    Aberdeen\n    1      10001266         France   780.34  Sales Executive    Aberdeen\n    2      10011532        Germany   181.68    Manual Worker      Alford\n    3      10011533        Germany   300.67    Manual Worker      Alford\n    4      10015830   Unclassified   228.70  Sales Executive     Macduff\n    ..          ...            ...      ...              ...         ...\n    995    10996176  United States   241.24     Professional  Glenrothes\n    996    10996177         Greece   343.23          Manager  Glenrothes\n    997    10996178  United States   636.22          Manager  Glenrothes\n    998    10996179  United States   356.21          Manager  Glenrothes\n    999    10996180  United States   438.20          Manager  Glenrothes\n\n    [1000 rows x 5 columns]\n\nYou can use a base selection to filter the records:\n\n.. code-block:: python\n\n    >>> sweden = dest == "29"\n    >>> sweden_datagrid = sweden.datagrid([urn, dest, cost, occupation, town])\n    >>> sweden_datagrid.to_df()\n        Booking URN Destination     Cost       Occupation           Town\n    0      10172319      Sweden  1201.81  Sales Executive         Bolton\n    1      10384970      Sweden   344.30          Manager     Chelmsford\n    2      10421011      Sweden   322.89  Sales Executive        Croydon\n    3      10425298      Sweden   880.02          Student  South Croydon\n    4      10479109      Sweden   172.91    Retail Worker       Nantwich\n    ..          ...         ...      ...              ...            ...\n    995    11471824      Sweden   118.76  Sales Executive    King\'s Lynn\n    996    11576762      Sweden   652.38    Public Sector        Redhill\n    997    11576764      Sweden   183.36    Public Sector        Redhill\n    998    11682962      Sweden  1166.38          Manager         London\n    999    11754655      Sweden   192.45  Sales Executive          Ascot\n\n    [1000 rows x 5 columns]\n\nYou can filter using a selection from a different table:\n\n.. code-block:: python\n\n    >>> manchester = households["Region"] == "13"\n    >>> manc_datagrid = manchester.datagrid(\n    ...     [urn, dest, cost, occupation, town], table=bookings\n    ... )\n    >>> manc_datagrid.to_df()\n        Booking URN    Destination     Cost       Occupation         Town\n    0      10172319         Sweden  1201.81  Sales Executive       Bolton\n    1      10172320  United States  1616.80  Sales Executive       Bolton\n    2      10173729         France   581.71          Student       Bolton\n    3      10173730         France  2224.70          Student       Bolton\n    4      10177047         France   686.53  Sales Executive       Bolton\n    ..          ...            ...      ...              ...          ...\n    995    11739340      Australia   316.60     Professional  Stalybridge\n    996    11739342   Unclassified   316.58  Sales Executive  Stalybridge\n    997    12087034         Greece  1305.66    Public Sector   Altrincham\n    998    12087035  United States   585.65    Public Sector   Altrincham\n    999    12087036      Australia   496.64    Public Sector   Altrincham\n\n    [1000 rows x 5 columns]\n\nCreating cubes\n==============\n\nYou can create a cube from a table:\n\n.. code-block:: python\n\n    >>> occupation = people["Occupation"]\n    >>> income = people["Income"]\n    >>> gender = people["Gender"]\n    >>> cube = people.cube([occupation, income, gender])\n\nConvert it to a Pandas DataFrame:\n\n.. code-block:: python\n\n    >>> df = cube.to_df()\n    >>> df.head(10)\n                                   People\n    Occupation    Income  Gender\n    Manual Worker <£10k   Female    15624\n                          Male       5321\n                          Unknown       5\n                  £10-20k Female    43051\n                          Male       5992\n                          Unknown      25\n                  £20-30k Female     1498\n                          Male        649\n                          Unknown      14\n                  £30-40k Female      675\n\nYou can pivot the dimensions to make it easier to read:\n\n.. code-block:: python\n\n    >>> df.unstack(level=0)\n                       People          ...\n    Occupation       Director Manager  ... Student Unemployed\n    Income   Gender                    ...\n    <£10k    Female      1279    4649  ...   28002      21385\n             Male         832    2926  ...   14296       8386\n             Unknown        4      16  ...      10        155\n    £10-20k  Female      4116   16665  ...   39462      17230\n             Male        2139    9123  ...   17917       4532\n             Unknown        9      47  ...      25        368\n    £100k+   Female         2       1  ...       2          0\n             Male           1       0  ...       3          0\n             Unknown        1       0  ...       1          0\n    £20-30k  Female      1267    6238  ...    6669       5747\n             Male        1050    5315  ...    5274       1345\n             Unknown        5      45  ...      22        236\n    £30-40k  Female      1591    6621  ...    5690       3117\n             Male        1940    9713  ...    6345       1049\n             Unknown       46     140  ...      63        519\n    £40-50k  Female       265     965  ...     587        262\n             Male         518    1800  ...     943        115\n             Unknown       22      58  ...      29        110\n    £50-60k  Female       336     806  ...     425        277\n             Male         607    1677  ...     692         69\n             Unknown       47      88  ...      64         89\n    £60-70k  Female        40     112  ...      54         58\n             Male          96     220  ...      95          8\n             Unknown       11      16  ...      17         17\n    £70-80k  Female        44      96  ...      42         27\n             Male         102     179  ...      63          5\n             Unknown       12      22  ...      15          5\n    £80-90k  Female        11      11  ...       3          0\n             Male          14      13  ...      16          0\n             Unknown        4       3  ...       5          0\n    £90-100k Female         1       0  ...       1          1\n             Male          11       7  ...       4          0\n             Unknown        3       6  ...       9          0\n\n    [33 rows x 10 columns]\n\nYou can use a base selection to filter the records:\n\n.. code-block:: python\n\n    >>> occupation = people["Occupation"]\n    >>> region = households["Region"]\n    >>> student = occupation == "4"\n    >>> student_cube = student.cube([occupation, dest, region])\n    >>> student_df = student_cube.to_df()\n    >>> student_df.head()\n                                                                     People\n    Occupation    Destination Region\n    Manual Worker Australia   North                                       0\n                              North West (Excluding Gtr Manchester)       0\n                              South East (Outside M25 )                   0\n                              South West                                  0\n                              East Midlands                               0\n\nSelecting only cells where ``Occupation`` is *Student*,\nand pivoting ``Destination`` dimension:\n\n.. code-block:: python\n\n    >>> student_df.loc["Student"].unstack(level=0)\n                                             People          ...\n    Destination                           Australia Denmark  ... Sweden United States\n    Region                                                   ...\n    Channel Islands                              46       1  ...     10            81\n    East Anglia                                 989       0  ...    109           905\n    East Midlands                              1956       0  ...    174          1762\n    Greater Manchester                         1197       1  ...    147          1089\n    North                                       959       2  ...    115           869\n    North West (Excluding Gtr Manchester)      1594       2  ...    177          1429\n    Northern Ireland                            467       0  ...     42           492\n    Scotland                                   2061       1  ...    224          1964\n    South East (Inside M25 )                   3935       0  ...    390          3580\n    South East (Outside M25 )                  6255       1  ...    608          5587\n    South West                                 2310       0  ...    182          2037\n    Wales                                       974       0  ...    122           860\n    West Midlands                              2643       0  ...    288          2362\n    Yorkshire and Humber                       2295       0  ...    249          2089\n\n    [14 rows x 19 columns]\n\nYou can use a selection from a different table to filter the records in the cube:\n\n.. code-block:: python\n\n    >>> manchester = region == "13"\n    >>> manc_cube = manchester.cube([occupation, dest, region], table=bookings)\n    >>> manc_cube.to_df()\n                                                                      Bookings\n    Occupation    Destination  Region\n    Manual Worker Australia    North                                         0\n                               North West (Excluding Gtr Manchester)         0\n                               South East (Outside M25 )                     0\n                               South West                                    0\n                               East Midlands                                 0\n                                                                        ...\n    Retired       South Africa Scotland                                      0\n                               Wales                                         0\n                               Northern Ireland                              0\n                               Greater Manchester                            0\n                               Channel Islands                               0\n\n    [2660 rows x 1 columns]\n\nYou can find the complete documentation\nincluding a more thorough `tutorial <https://help.apteco.com/python/tutorial.html>`_\non the `Apteco website <https://help.apteco.com/python/index.html>`_.',
    'author': 'Apteco Ltd',
    'author_email': 'support@apteco.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://help.apteco.com/python/index.html',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
